from _thread import get_ident

import itertools

from PyQt5.QtCore import pyqtSignal, QObject, QReadWriteLock, QTimer, pyqtSlot

from .element import *
from .errors import ConnectError, GeneralException
from .serialization import ComplexJsonEncoder, ComplexJsonDecoder
from ..version import __version__


class ReadLocker:
    def __init__(self, parent):
        assert isinstance(parent, DiagramLock)
        self.parent = parent

    def __enter__(self):
        if not self.parent.tryLockForRead():
            if self.parent.owner == get_ident():
                self.parent.lockForWrite()
            else:
                self.parent.lockForRead()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.parent.unlock()


class WriteLocker:
    def __init__(self, parent):
        assert isinstance(parent, DiagramLock)
        self.parent = parent
        self.count = 0

    def __enter__(self):
        self.parent.lockForWrite()
        self.count += 1
        self.parent.owner = get_ident()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.count -= 1
        if not self.count:
            self.parent.owner = None
        self.parent.unlock()


class DiagramLock(QReadWriteLock):
    def __init__(self):
        QReadWriteLock.__init__(self, QReadWriteLock.Recursive)
        self.writer = WriteLocker(self)
        self.reader = ReadLocker(self)
        self.owner = None


class Diagram(QObject):
    """The processing diagram"""

    # todo: if the singals and QObject dependency are removed, Workarea or ScrolledWorkarea could derive from
    # Diagram, which would possibly make thing simpler
    element_added = pyqtSignal(Element, tuple)
    element_deleted = pyqtSignal(Element)
    connection_created = pyqtSignal(Output, Input)
    connection_deleted = pyqtSignal(Output, Input)

    # global diagram lock for connecting and disconnecting elements
    diagram_lock = DiagramLock()

    def __init__(self):
        super(Diagram, self).__init__()
        self.elements = set()
        self.connections = []
        self.painter = None
        self.zoom_level = 1.0

    def clear(self):
        for e in list(self.elements):
            self.delete_element(e)

    def add_element(self, e, position):
        if not self.painter:
            raise GeneralException("Elements cannot be added to Diagram until the painter is set")
        e.diagram = self
        self.elements.add(e)
        self.element_added.emit(e, position)

    def delete_element(self, e):
        # todo: if we remove element, to which other elements try to access, we gonna have trouble
        with self.diagram_lock.writer:
            to_connect = []
            if len(e.inputs) == 1 and len(e.outputs) == 1:
                input_ = list(e.inputs.values())[0]
                output = list(e.outputs.values())[0]
                pre = input_.connected_from
                post = output.connected_to
                if len(pre) > 0 and len(post) > 0:
                    for i, o in itertools.product(pre, post):
                        # todo: it's not optimal when 'i' is multiple but 'o' is not
                        #self.connect_io(i, o)
                        to_connect.append((i, o))
        for p in list(e.parameters.values()):
            p.disconnect_all_children()
        for io in list(e.outputs.values()) + list(e.inputs.values()):
            io.disconnect_all()
            self.delete_connections_with_connector(io)
        e.delete()
        self.elements.remove(e)
        self.element_deleted.emit(e)
        for i, o in to_connect:
            self.connect_io(i, o)

    def delete_connections_with_connector(self, io):
        for connection in [c for c in self.connections if io in c]:
            self.connections.remove(connection)

    def connect_io(self, o1, o2):
        if o1 is o2: return
        with self.diagram_lock.writer:
            if isinstance(o1, Input):
                input_ = o1
                output = o2
            else:
                input_ = o2
                output = o1
            if Diagram.makes_loop(output, input_):
                print("WARNING Connection loop: {}:{} -> {}:{}".format(output.parent.name, output.name, input_.parent.name, input_.name))
                # raise ConnectError("This connection would create an infinite loop!")
                return
            input_.connect(output)
            output.connect(input_)
            if output.desequencing:
                output.hook.actualize_outputs()
        self.connections.append((output, input_))
        self.connection_created.emit(output, input_)

    def disconnect_io(self, o1, o2):
        if o1 is o2: return
        with self.diagram_lock.writer:
            if isinstance(o1, Input):
                input_ = o1
                output = o2
            else:
                input_ = o2
                output = o1
            output.disconnect(input_)
            input_.disconnect(output)
            if output.desequencing:
                output.hook.actualize_outputs()
        self.connections.remove((output, input_))

    def notify_disconnect(self, output, input_):
        self.connection_deleted.emit(output, input_)

    @staticmethod
    def makes_loop(output, input_):
        # DFS
        with Diagram.diagram_lock.writer:
            q = [input_.parent]
            c = set()
            while q:
                act = q.pop()
                if act in c:
                    continue
                if act is output.parent:
                    return True
                c.add(act)
                for i in list(act.outputs.values()):
                    for o in i.connected_to:
                        if o.parent is output.parent:
                            return True
                        if o.parent not in c:
                            q.append(o.parent)
        return False

    def set_painter(self, painter):
        self.painter = painter

    def save_to_json(self):
        return ComplexJsonEncoder(indent=2, sort_keys=True).encode(self)

    def load_from_json(self, ascii_data):
        if not self.painter:
            raise GeneralException("Diagram cannot be filled with data until the painter is set")
        ComplexJsonDecoder(self).decode(ascii_data)
        QTimer.singleShot(100, self.update_previews)

    @pyqtSlot()
    def update_previews(self):
        for e in self.elements:
            e.preview.force_update()

    def to_json(self):
        elements = {}
        elements_orders = {}
        wires = {}

        for e in self.elements:
            e_order = self.painter.element_z_index(e)
            # z-indexes are unique, but just in case:
            while e_order in elements:
                e_order += 1
            elements[e_order] = e
            elements_orders[e] = e_order

        for c_order, c in enumerate(self.connections):
            wire = {
                "from_element": elements_orders[c[0].parent],
                "from_output": c[0].id,
                "to_element": elements_orders[c[1].parent],
                "to_input": c[1].id
            }
            wires[c_order] = wire

        param_ids = {}
        for e_id, e in sorted(elements.items()):
            for par in e.parameters.values():
                param_ids[par] = len(param_ids)

        connected_params = []
        for par, par_id in param_ids.items():
            assert isinstance(par, Parameter)
            for to in par.children:
                assert isinstance(to, Parameter)
                connected_params.append({
                    "from": par_id,
                    "to": param_ids[to]
                })
                # connected_params.append([par_id, param_ids[to]])

        filetype = "CV-Lab diagram save file. See: https://github.com/cvlab-ai/cvlab "

        return {"_type": "diagram", "elements": elements, "wires": wires, "params": connected_params,
                "zoom_level": self.zoom_level, "_version": __version__, "_filetype": filetype}

    def from_json(self, data):
        #TODO: catch json parsing errors and present proper message
        elements = {}
        sorted_orders = sorted(map(int, data["elements"]))  # sorting is important for preserving z-index

        for e_order in sorted_orders:
            e = data["elements"][str(e_order)]
            self.add_element(e, (e.pos().x(), e.pos().y()))
            elements[e_order] = e

        sorted_orders = sorted(map(int, data["wires"]))     # sorting is important for preserving connections order
        for c_order in sorted_orders:
            connection = data["wires"][str(c_order)]
            from_e_id = connection["from_element"]
            from_o_id = connection["from_output"]
            to_e_id = connection["to_element"]
            to_i_id = connection["to_input"]

            # todo: remove it. It's only a workaround for old Forwarder element
            if elements[from_e_id].name == "Forwarder": from_o_id = "output"
            if elements[to_e_id].name == "Forwarder": to_i_id = "input"

            a = elements[from_e_id].outputs[from_o_id]
            b = elements[to_e_id].inputs[to_i_id]
            self.connect_io(a, b)

        if 'params' in data:
            param_ids = {}
            for e_id, e in sorted(elements.items()):
                for par in e.parameters.values():
                    param_ids[len(param_ids)] = par

            for con in data['params']:
                from_ = con['from']
                to_ = con['to']
                param_ids[from_].connect_child(param_ids[to_])

        self.zoom_level = data.get("zoom_level", 1.0)



import sys
import importlib
import os
from collections import defaultdict

import re

ignored_modules = ["sample", "testing"]


registered_elements = defaultdict(list)  # package -> [elements]
sort_keys = defaultdict(lambda: 999)
all_elements = {}


def element_name(name):
    name = re.match(r"(cvlab[^.]*\.)?(diagram\.)?(elements\.)?(experimental\.|testing\.|custom\.)?(.+)", name).group(5)
    return name


def register_elements(package, elements, sort_key=999):
    sort_keys[package] = min(sort_key, sort_keys[package])
    for element in elements:
        element_package = getattr(element, "package", None) or package
        module = element.__module__
        classname = element.__name__
        name = module + "." + classname
        registered_elements[element_package].append(element)
        all_elements[element_name(name)] = element
        sort_keys[element_package] = min(sort_keys[element_package], sort_key + 100)


def register_elements_auto(module_name, module_locals, package, sort_key=999):
    elements = [cls for cls in module_locals.values() if isinstance(cls, type) and cls.__module__ == module_name and hasattr(cls, "name")]
    return register_elements(package, elements, sort_key)


def get_sorted_elements():
    return sorted(registered_elements.items(), key=lambda kv: sort_keys[kv[0]])


def get_element_fallback(name):
    class_name = name.split(".")[-1]
    for element in all_elements.values():
        if element.__name__ == class_name:
            print("WARN: Loading fallback element. Requested name: {name}. Returned class: {element}".format(**locals()))
            return element
    raise Exception("Cannot find element " + name)


def get_element(name):
    name = element_name(name)
    element = all_elements.get(name, None)
    if not element: element = get_element_fallback(name)
    return element


def available_modules(path):
    modules = []
    dir = os.path.realpath(path)
    if not os.path.isdir(dir): dir = os.path.dirname(dir)
    for entry in os.listdir(dir):
        if entry == '__init__.py':
            continue
        if entry[-3:] == ".py":
            entry = entry[:-3]
        elif not os.path.isdir(dir + "/" + entry):
            continue
        if os.path.isdir(dir + "/" + entry) and not os.path.exists(dir + "/" + entry + "/__init__.py"):
            continue
        modules.append(entry)
    return modules


def load_modules(modules, package):
    for module in modules:
        if module in ignored_modules:
            print("Ignoring module:", module)
            continue
        try:
            print("Loading module:", module)
            importlib.import_module("." + module, package)
        except BaseException as e:
            print("ERROR during loading module {module}: {e}".format(**locals()))
            # if __debug__:
            #     traceback.print_exc()


def load_plugins():
    for path in sys.path:
        if not os.path.isdir(path): continue
        for module in os.listdir(path):
            if not module.startswith("cvlab_"): continue
            module_path = path + "/" + module
            if os.path.isdir(module_path) and not os.path.isfile(module_path + "/__init__.py"): continue
            if module in sys.modules: continue
            try:
                print("Loading plugin:", module)
                importlib.import_module(module)
            except Exception as e:
                print("Error while loading module", module, ":", e)


def load_auto(path):
    modules = available_modules(path)
    package = path.replace("\\","/")
    package = package[package.index("cvlab"):]
    package = re.sub(r"/__init__\.py.*", "", package)
    package = package.replace("/",".")
    package = package.replace("cvlab.cvlab","cvlab")
    load_modules(modules, package)


load_auto(__file__)
load_plugins()



import itertools

from PyQt5.QtCore import Qt

from .base import *


class MatrixGenerator(InputElement):
    name = "Matrix generator"
    comment = "Generates selected matrix"

    def get_attributes(self):
        return [], [Output("output")], [
            ComboboxParameter("type", {"Rectangle": 1, "Circle": 2, "Rectangle [mono]": 3, "Circle [mono]": 4}),
            IntParameter("size", min_=1)]

    def process_inputs(self, inputs, outputs, parameters):
        size = parameters["size"]
        output = None

        if parameters["type"] == 1: #rectangle
            output = np.empty((size, size, 3), np.uint8)
            output.fill(255)

        elif parameters["type"] == 2: #circle
            output = np.zeros((size, size, 3), np.uint8)
            cv.circle(output, (size // 2, size // 2), size // 2, (255, 255, 255), -1)

        elif parameters["type"] == 3: #rectangle
            output = np.empty((size, size), np.uint8)
            output.fill(255)

        elif parameters["type"] == 4: #circle
            output = np.zeros((size, size), np.uint8)
            cv.circle(output, (size // 2, size // 2), size // 2, 255, -1)

        if output is None: raise ProcessingError("Wrong matrix type")

        outputs["output"] = Data(output)


class MatrixEditor(InputElement):
    name = "Matrix editor"
    comment = "Allows to create a matrix pixel by pixel"

    def __init__(self):
        super(MatrixEditor, self).__init__()
        self.last_parameters = {}
        self.edit_widget = QWidget()
        self.matrix = None
        self.layout().addWidget(self.edit_widget)
        self.edit_widget.setLayout(QGridLayout())

    def get_attributes(self):
        return [], [Output("output")], [
            ComboboxParameter("type", {"8-bit unsigned":np.uint8, "8-bit signed":np.int8, "32-bit float": np.float32}),
            SizeParameter("size"),
            FloatParameter("step", value=1)]

    def process_inputs(self, inputs, outputs, parameters):
        if parameters != self.last_parameters:
            self.last_parameters = parameters.copy()
            self.recreate_editor()
        outputs["output"] = Data(self.matrix+0)

    class Button(QLabel):
        pixel_size = (64, 64)

        def __init__(self, element, x, y, step):
            QLabel.__init__(self)
            self.element = element
            self.matrix_x = x
            self.matrix_y = y
            self.step = step
            self.setFixedSize(self.pixel_size[0], self.pixel_size[1])
            self.setAlignment(Qt.AlignCenter | Qt.AlignVCenter)

        def mousePressEvent(self, event):
            if event.button() == Qt.LeftButton:
                delta = self.step
            elif event.button() == Qt.RightButton:
                delta = -self.step
            else:
                event.ignore()
                return
            event.accept()
            self.add_value(delta)

        def mouseReleaseEvent(self, event):
            if event.button() in (Qt.LeftButton, Qt.RightButton):
                event.accept()
            else:
                event.ignore()

        def wheelEvent(self, event):
            event.accept()
            delta = self.step if event.angleDelta().y() > 0 else -self.step
            self.add_value(delta)

        def add_value(self, delta):
            self.element.matrix[self.matrix_y,self.matrix_x] += delta
            self.setText(str(self.element.matrix[self.matrix_y,self.matrix_x]))
            self.element.recalculate(False, False, False, force_units_recalc=True)


    def recreate_editor(self):
        QWidget().setLayout(self.edit_widget.layout())
        layout = QGridLayout()
        self.edit_widget.setLayout(layout)
        w, h = self.last_parameters["size"]
        dtype = self.last_parameters["type"]
        step = self.last_parameters["step"]
        self.matrix = np.zeros((h, w), dtype=dtype)
        layout.setSpacing(0)
        for x, y in itertools.product(range(w), range(h)):
            pixel = self.Button(self, x, y, step)
            pixel.setText("0")
            layout.addWidget(pixel, y, x)


class MatrixEditor2(InputGuiElement, ThreadedElement):
    name = "Matrix editor 2"
    comment = "Allows to create a matrix pixel by pixel"

    def get_attributes(self):
        return [], [Output("output")], [MatrixParameter("matrix")]

    def process_inputs(self, inputs, outputs, parameters):
        outputs["output"] = Data(parameters["matrix"])


class OpenCVGetStructuringElement(InputElement):
    name = "Structuring Element"
    comment = "Returns structuring element"

    def get_attributes(self):
        return [], \
               [Output("output")], \
               [
                    ComboboxParameter("Element type", values=OrderedDict([
                        ("MORPH_RECT", cv.MORPH_RECT),
                        ("MORPH_ELLIPSE", cv.MORPH_ELLIPSE),
                        ("MORPH_CROSS", cv.MORPH_CROSS)
                    ])),
                    IntParameter("x element size", "Width of structuring element", 5, min_=1, max_=255, step=1),
                    IntParameter("y element size", "Height of structuring element", 5, min_=1, max_=255, step=1),
               ]

    def process_inputs(self, inputs, outputs, parameters):
        element_type = parameters["Element type"]
        x_element_size = parameters["x element size"]
        y_element_size = parameters["y element size"]

        element = cv.getStructuringElement(element_type, (x_element_size, y_element_size))
        outputs["output"] = Data(element)


register_elements_auto(__name__, locals(), "Data generation", 7)



import time

from .hooks import *
from .exceptions import *
from .processing_time import ProcessingTimeInfo
from .core_element import CoreElement


class ThreadedElement(CoreElement):
    """Base class for all threaded elements"""

    def __init__(self):
        super(ThreadedElement, self).__init__()
        self.state = self.STATE_UNSET
        self._do_abort = False
        self._do_break = False
        self._notifier = Notifier()
        self._worker = threading.Thread(target=self.work)
        self._worker.daemon = True
        self._worker.name = "Worker for '" + self.__class__.__name__ + "'"
        self._worker.start()
        self.processing_time_info = None

    def recalculate(self, refresh_parameters, refresh_structure, force_break, force_units_recalc=False):
        self.structure_changed |= refresh_structure
        self.parameters_changed |= refresh_parameters
        self._do_break |= force_break
        self._notifier.notify()

    def may_interrupt(self):
        if self._do_break:
            raise InterruptException()

    def delete(self):
        self._do_abort = True
        self._do_break = True
        self._notifier.notify()
        self._worker.join()
        CoreElement.delete(self)

    def work(self):
        self.set_state(self.STATE_UNSET)
        while True:
            self._notifier.wait()
            if self._do_abort: break
            try:
                self.set_state(self.STATE_BUSY)
                start = time.clock()
                self._do_break = False
                self.process()
                self.may_interrupt()
                end = time.clock()
                previous_time_infos = self.get_previous_time_infos()
                self.processing_time_info = ProcessingTimeInfo(start, end, len(self.units), previous_time_infos)
                self.set_state(self.STATE_READY)
            except (InterruptException, ProcessingBreak):
                pass
            except Exception as e:
                self.set_state(self.STATE_ERROR, e)
        self.set_state(self.STATE_UNSET)

    def get_previous_time_infos(self):
        time_infos = []
        for connector in self.inputs.values():
            for inpt in connector.connected_from:
                if hasattr(inpt.parent, "processing_time_info"):    # fixme: ugly hack...
                    time_infos.append(inpt.parent.processing_time_info)
        return time_infos





import os
import sys
from threading import Thread

from xmlrpc.client import ServerProxy

from pip._vendor.packaging.version import parse as parse_version

from ..version import package_name, __version__


class Updater:
    pypi_url = 'https://pypi.python.org/pypi'

    def check(self):
        act_version = parse_version(__version__)

        pypi = ServerProxy(self.pypi_url)
        versions = pypi.package_releases(package_name)

        if not versions: raise Exception("There are no available releases on PyPI")

        versions = map(parse_version, versions)
        newest_version = max(versions)

        if newest_version > act_version:
            return True, newest_version
        else:
            return False, act_version

    def check_async(self, callback):
        thread = Thread(name="Updater thread", target=self._check_async, args=[callback])
        thread.setDaemon(True)
        thread.start()

    def _check_async(self, callback):
        try:
            can_update, version = self.check()
            callback(can_update, str(version))
        except Exception as e:
            print("WARNING: Cannot get update information:", e)

    def update_command(self):
        cvlab_dir = os.path.normpath(os.path.abspath(__file__) + "/../..")

        # inside git repo?
        git_path = cvlab_dir + "/../.git"
        if os.path.exists(git_path):
            repo = os.path.normpath(cvlab_dir + "/..")
            return "git -C {repo} pull".format(**locals())

        # inside virtualenv?
        if hasattr(sys, 'real_prefix'):
            pip = os.path.dirname(sys.executable) + "/pip3"
            pip = os.path.normpath(pip)
            return "{pip} install --upgrade cvlab".format(**locals())

        # inside user's pip directory?
        if os.access(__file__, os.W_OK) and ".local" in __file__:
            return "pip3 install --user --upgrade cvlab"

        # inside global pip directory, needs sudo?
        if not os.access(__file__, os.W_OK):
            return "sudo -H pip3 install --upgrade cvlab"

        # general
        return "pip3 install --upgrade cvlab"




import json


class ComplexJsonEncoder(json.JSONEncoder):
    def default(self, o):
        if hasattr(o, 'to_json'):
                return o.to_json()
        else:
            raise TypeError('Object of type %s with value of %s is not JSON serializable' % (type(o), repr(o)))


class ComplexJsonDecoder(json.JSONDecoder):
    def __init__(self, diagram):
        json.JSONDecoder.__init__(self, object_hook=self.dict_to_object)
        self.diagram_instance = diagram

    def dict_to_object(self, d):
        if "_type" in d and d["_type"] == "element":
            class_name = d["class"]
            module_name = d["module"]
            from .elements import get_element
            element = get_element(module_name + "." + class_name)()
            element.from_json(d)
            return element
        if "_type" in d and d["_type"] == "diagram":
            self.diagram_instance.from_json(d)
            return self.diagram_instance
        return d



import os

from PyQt5.QtCore import Qt, pyqtSignal, QTimer
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from ..core.update import Updater, parse_version
from .diagram_manager import DiagramManager
from .menubar import MenuBar
from .tabs_container import TabsContainer
from .toolbox import Toolbox
from .config import ConfigWrapper, UPDATES_SECTION, UPDATE_DONT_REMIND_VERSION
from .styles import StyleManager


LOAD_LAST_DIAGRAMS = True
LOAD_DEFAULT_DIAGRAM = True
DEFAULT_DIAGRAM_PATH = "default.cvlab"
ICON_PATH = os.path.dirname(__file__) + "/../images/icon.png"
FONT_PATH = os.path.dirname(__file__) + "/../styles/fonts/Carlito-Regular.ttf"


class MainWindow(QMainWindow):
    update_available = pyqtSignal(bool, str)

    def __init__(self, application):
        super(MainWindow, self).__init__()

        assert isinstance(application, QApplication)

        self.application = application
        self.settings = ConfigWrapper.get_settings()
        self.style_name = None
        self.resize(1000, 600)
        self.setWindowTitle('CV Lab')

        icon = QIcon(ICON_PATH)
        application.setWindowIcon(icon)
        QFontDatabase.addApplicationFont(FONT_PATH)
        self.style_manager = StyleManager(self)

        self.toolbox = Toolbox()
        self.tabs_container = TabsContainer()
        self.diagram_manager = DiagramManager(self.tabs_container, self.style_manager)
        self.tabs_container.diagram_manager = self.diagram_manager
        MenuBar(self)

        layout = QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.toolbox, 0, 0)
        layout.addWidget(self.tabs_container, 0, 1)

        layout.setColumnMinimumWidth(0, 220)
        layout.setColumnStretch(0, 1)
        layout.setColumnStretch(1, 6)

        central = QWidget()
        central.setLayout(layout)
        self.setCentralWidget(central)
        self.showMaximized()
        self.raise_()   # bring to front on a Mac

        QTimer.singleShot(100, self.load_diagrams)

        # Prevent automatic focus on toolbox search field
        self.toolbox.filter_input.clearFocus()

        # check for updates
        self.updater = Updater()
        self.update_available.connect(self.show_update_info)
        self.updater.check_async(self.update_available.emit)

    def load_diagrams(self):
        if LOAD_LAST_DIAGRAMS:
            self.diagram_manager.open_from_settings(self.settings)
            if self.diagram_manager.count() == 0:
                if LOAD_LAST_DIAGRAMS:
                    self.diagram_manager.open_diagram_from_path(DEFAULT_DIAGRAM_PATH)
                else:
                    self.diagram_manager.open_diagram()
        elif LOAD_DEFAULT_DIAGRAM:
            self.diagram_manager.open_diagram_from_path(DEFAULT_DIAGRAM_PATH)
        else:
            self.diagram_manager.open_diagram()

    def keyPressEvent(self, event):
        key = event.key()
        if key == Qt.Key_Escape:
            self.toolbox.filter_input.setFocus()
            self.toolbox.filter_input.selectAll()
        if (key == Qt.Key_Tab and (event.modifiers() & Qt.ControlModifier)) or \
                (key == Qt.Key_Backtab and (event.modifiers() & Qt.ControlModifier)):
            self.tabs_container.keyPressEvent(event)

    def closeEvent(self, event):
        QMainWindow.closeEvent(self, event)
        self.diagram_manager.save_to_settings(self.settings)
        self.diagram_manager.close_all_diagrams()
        # Close any remaining windows:
        for window in [x for x in QApplication.topLevelWidgets() if x.isWindow() and x != self and x.parent() is None and x.isVisible()]:
            window.close()

    def show_update_info(self, can_update, newest_version):
        from ..version import __version__
        newest_version = str(newest_version)
        if can_update:
            act_version = parse_version(__version__)
            newest_version = parse_version(newest_version)
            dont_remind_version = parse_version(self.settings.get_with_default(UPDATES_SECTION, UPDATE_DONT_REMIND_VERSION))

            command = self.updater.update_command()
            self.application.clipboard().setText(command)

            message = """\
There is a new version of CV Lab.

Your version: {__version__}
Newest version: {newest_version}

To update you may run:
{command}

(command copied to clipboard)
""".format(**locals())

            if newest_version <= dont_remind_version:
                print("--- UPDATE INFORMATION ---")
                print(message)
            else:
                mb = QMessageBox(QMessageBox.Information,"Update information", message)
                mb.addButton(QPushButton("OK"), QMessageBox.AcceptRole)
                ignore = QPushButton("Ignore this update")
                mb.addButton(ignore, QMessageBox.DestructiveRole)
                mb.exec_()
                if mb.clickedButton() == ignore:
                    self.settings.set(UPDATES_SECTION, UPDATE_DONT_REMIND_VERSION, str(newest_version))
        else:
            print("You're using newest version of CV Lab ({__version__}).".format(**locals()))




from cvlab.view.config import ConfigWrapper, ELEMENTS_SECTION, EXPERIMENTAL_ELEMENTS
if ConfigWrapper.get_settings().get_with_default(ELEMENTS_SECTION, EXPERIMENTAL_ELEMENTS) == "True":
    from cvlab.diagram.elements import load_auto, ignored_modules
    ignored_modules += ["sample"]
    load_auto(__file__)


from copy import *

from .base import *


class PlusOperator(MultiInputOneOutputElement):
    name = "Plus operator"
    comment = "Adds all input images"

    def process_inputs(self, inputs, outputs, parameters):
        output = Data()
        for input in inputs.values():
            if output.type() == output.NONE:
                output.value = copy(input.value)
            else:
                cv.add(output.value, input.value, output.value)
            self.may_interrupt()
        outputs["output"] = output


class MinusOperator(NormalElement):
    name = "Minus operator"
    comment = "Subtracts second from first image"

    def get_attributes(self):
        return [Input("from"), Input("what")], [Output("output")], []

    def process_inputs(self, inputs, outputs, parameters):
        outputs["output"] = Data(cv.subtract(inputs["from"].value, inputs["what"].value))


class AbsDiffOperator(NormalElement):
    name = "Difference operator"
    comment = "Calculates absolute difference between images"

    def get_attributes(self):
        return [Input("1", "Input 1"), Input("2", "Input 2")], [Output("output")], []

    def process_inputs(self, inputs, outputs, parameters):
        outputs["output"] = Data(cv.absdiff(inputs["1"].value, inputs["2"].value))


class AverageOperator(MultiInputOneOutputElement):
    name = "Average operator"
    comment = "Averages all input images"

    def process_inputs(self, inputs, outputs, parameters):
        temp = None
        for input_ in inputs.values():
            if temp is None:
                temp = np.float32(input_.value)
            else:
                temp += np.float32(input_.value)
        outputs["output"] = Data(np.uint8(temp // len(inputs)))


class MaxOperator(MultiInputOneOutputElement):
    name = "Maximum operator"
    comment = "Output image contains maximum pixel values"

    def process_inputs(self, inputs, outputs, parameters):
        temp = None
        for input_ in inputs.values():
            if temp is None:
                temp = copy(input_.value)
            else:
                temp = np.maximum(temp, input_.value)
        outputs["output"] = Data(temp)


class MinOperator(MultiInputOneOutputElement):
    name = "Minimum operator"
    comment = "Output image contains minimum pixel values"

    def process_inputs(self, inputs, outputs, parameters):
        temp = None
        for input_ in inputs.values():
            if temp is None:
                temp = copy(input_.value)
            else:
                temp = np.minimum(temp, input_.value)
        outputs["output"] = Data(temp)


class InvertOperator(NormalElement):
    name = "Invertion operator"
    comment = "Inverts pixel values"

    def get_attributes(self):
        return [Input("input")], [Output("output")], []

    def process_inputs(self, inputs, outputs, parameters):
        outputs["output"] = Data(-inputs["input"].value - 1)


class ScalarMultiplyOperator(NormalElement):
    name = "Scalar multiply"
    comment = "Multiplies matrice by scalar"

    def get_attributes(self):
        return [Input("input")], [Output("output")], [FloatParameter("factor", min_=0, max_=10)]

    def process_inputs(self, inputs, outputs, parameters):
        input_ = inputs["input"].value
        factor = parameters["factor"]
        if len(input_.shape) == 2:
            output = cv.multiply(input_, factor)
        else:
            output = cv.multiply(input_, (factor, factor, factor, factor))
        outputs["output"] = Data(output)


class ScalarAddOperator(NormalElement):
    name = "Scalar add"
    comment = "Adds constant value to each matrix element"

    def get_attributes(self):
        return [Input("input")], [Output("output")], [FloatParameter("value", min_=-255, max_=255)]

    def process_inputs(self, inputs, outputs, parameters):
        input_ = inputs["input"].value
        value = parameters["value"]
        if len(input_.shape) == 2:
            output = cv.add(input_, value)
        else:
            output = cv.add(input_, (value, value, value, value))
        outputs["output"] = Data(output)


register_elements_auto(__name__, locals(), "Basic operators", 4)



from cvlab.diagram.elements.base import *

from .ml import Trainable


class SvmTrain(Trainable):
    name = "SvmTrain"
    comment = "Support Vector Machine classifier training"

    output_type = Trainable.TYPE_CLASSES
    classes_count = 2

    def __init__(self):
        super(SvmTrain, self).__init__()
        self.svm = None
        self.train_data = None
        self.params_ = None

    def get_attributes(self):
        return [Input("train data"),
                Input("responses")], \
               [Output("model"),
                Output("log")], \
               [
                   ComboboxParameter("svm_type", values=OrderedDict([
                       ("C_SVC", cv.SVM_C_SVC),
                       ("NU_SVC", cv.SVM_NU_SVC),
                       ("ONE_CLASS", cv.SVM_ONE_CLASS),
                       ("EPS_SVR", cv.SVM_EPS_SVR),
                       ("NU_SVR", cv.SVM_NU_SVR)
                    ])),
                   ComboboxParameter("kernel_type", values=OrderedDict([
                       ("LINEAR", cv.SVM_LINEAR),
                       ("POLY", cv.SVM_POLY),
                       ("RBF", cv.SVM_RBF),
                       ("SIGMOID", cv.SVM_SIGMOID)
                   ]), default_value_idx=2),
                   FloatParameter("degree", value=0.000001),
                   FloatParameter("gamma", value=1.0),
                   FloatParameter("coef0", value=0.0),
                   FloatParameter("C", value=1.0),
                   FloatParameter("nu", value=0.000001),
                   FloatParameter("p", value=0.0),
                   IntParameter("max_iter", value=1000),
                   FloatParameter("epsilon", value=0.000001, min_=0.0000001, max_=1.0),
                   IntParameter("CV_k", value=1, min_=0, max_=100)
            ]

    def process_inputs(self, inputs, outputs, parameters):
        train_data = inputs["train data"].value
        responses = inputs["responses"].value
        print("SVM process_intputs", train_data.shape, responses.shape)
        self.params_ = parameters
        sample_weights = self.get_sample_weights(responses)
        log = self.train_and_evaluate(train_data, responses, sample_weights, parameters["CV_k"])
        outputs["model"] = Data(self.svm)
        outputs["log"] = Data(log)

    def train(self, train_data, responses, sample_weights):
        print("Training SVM with", train_data.shape[0], "samples...")
        params = self.params_
        term_criteria = (cv.TERM_CRITERIA_MAX_ITER + cv.TERM_CRITERIA_EPS, params["max_iter"], params["epsilon"])
        svm_params = dict(
            svm_type=params["svm_type"],
            kernel_type=params["kernel_type"],
            degree=params["degree"],
            gamma=params["gamma"],
            coef0=params["coef0"],
            C=params["C"],
            nu=params["nu"],
            p=params["p"],
            term_crit=term_criteria)

        self.svm = cv.SVM()
        self.svm.train(train_data, responses, params=svm_params)

    def predict(self, v):
        return self.svm.predict(v)

    def evaluate_training(self, train_data, responses):
        ok_count = 0
        i = 0
        for v in train_data:
            if self.svm.predict(v) == responses[i]:
                ok_count += 1
            i += 1
        return "Train data result:\n   acc:\t" + str(float(ok_count) / i) + " (" + str(ok_count) + "/" + str(i) + ")"


class SvmTrainNoParams(Trainable):
    name = "SvmTrainNoParams"
    comment = "Support Vector Machine classifier"

    output_type = Trainable.TYPE_CLASSES
    classes_count = 2

    def __init__(self):
        super(SvmTrainNoParams, self).__init__()
        self.svm = None

    def get_attributes(self):
        return [Input("train data"),
                Input("responses")], \
               [Output("model"),
                Output("log")], \
               [IntParameter("CV_k", value=1, min_=0, max_=100)]

    def process_inputs(self, inputs, outputs, parameters):
        train_data = inputs["train data"].value
        responses = inputs["responses"].value
        sample_weights = self.get_sample_weights(responses)
        log = self.train_and_evaluate(train_data, responses, sample_weights, parameters["CV_k"])
        outputs["model"] = Data(self.svm)
        outputs["log"] = Data(log)

    def train(self, train_data, responses, sample_weights=None):
        self.svm = cv.SVM(train_data, responses)

    def predict(self, v):
        return self.svm.predict(v)


class SvmPredict(FunctionGuiElement, ThreadedElement):
    name = "SvmPredict"
    comment = "Support Vector Machine classifier prediction"

    def __init__(self):
        super(SvmPredict, self).__init__()
        self.svm = None

    def get_attributes(self):
        return [Input("test data"),
                Input("model")], \
               [Output("responses")], \
               []

    def process_inputs(self, inputs, outputs, parameters):
        self.svm = inputs["model"].value
        test_data = inputs["test data"].value
        predictions = []
        for v in test_data:
            result = self.svm.predict(v)
            predictions.append(result)
        result = np.array(predictions)
        outputs["responses"] = Data(result)


class AnnMlpTrain(Trainable):
    name = "ANN MLP Train"
    comment = "Artificial Neural Network - Multi layer perceptron - classifier training"

    output_type = Trainable.TYPE_REAL_CENTER_0_WITH_MIDDLE
    classes_count = 2

    def __init__(self):
        super(AnnMlpTrain, self).__init__()
        self.ann = None
        self.train_data = None
        self.params_ = None

    def get_attributes(self):
        return [Input("train data"),
                Input("responses")], \
               [Output("model"),
                Output("log")], \
               [
                   TextParameter("layer_sizes", value="5 3 1", window_title="Neural network layer sizes editor",
                                 window_content="Layers sizes from input to output, separated by spaces:"),
                   ComboboxParameter("activate_func", values=OrderedDict([
                       ("IDENTITY", cv.ml.ANN_MLP_IDENTITY),
                       ("SIGMOID_SYM", cv.ml.ANN_MLP_SIGMOID_SYM),
                       ("GAUSSIAN", cv.ml.ANN_MLP_GAUSSIAN),
                   ]), default_value_idx=1),
                   ComboboxParameter("train_method", values=OrderedDict([
                       ("BACKPROP", cv.ml.ANN_MLP_BACKPROP),
                       ("RPROP", cv.ml.ANN_MLP_RPROP),
                   ]), default_value_idx=1),
                   ComboboxParameter("scaling", values=OrderedDict([
                       ("FULL SCALING", 0),
                       ("NO INPUT SCALE", cv.ml.ANN_MLP_NO_INPUT_SCALE),
                       ("NO OUTPUT SCALE", cv.ml.ANN_MLP_NO_OUTPUT_SCALE),
                       ("NO SCALING", cv.ml.ANN_MLP_NO_INPUT_SCALE + cv.ml.ANN_MLP_NO_OUTPUT_SCALE),
                   ])),
                   # term criteria
                   IntParameter("max_iter", value=10, min_=1, max_=10000),
                   FloatParameter("epsilon", value=0.01, min_=0, max_=1),
                   # backprop params:
                   FloatParameter("dw_scale", value=0.1, min_=0, max_=1),
                   FloatParameter("moment_scale", value=0.1, min_=0, max_=2),
                   # rprop params
                   FloatParameter("dw0", value=0.1),
                   FloatParameter("dw_min", value=0.000001, min_=0),
                   # cross validation
                   IntParameter("CV_k", value=1, min_=0, max_=100),
                   # retraining
                   ComboboxParameter("train_mode", values=OrderedDict([
                       ("RETRAIN", 0),
                       ("UPDATE", cv.ml.ANN_MLP_UPDATE_WEIGHTS)
                   ])),
                   ButtonParameter("retrain_button", self.retrain, "Retrain ANN"),
                   # saving/loading
                   ButtonParameter("save", self.save, "Save net to file"),
               ]

    def process_inputs(self, inputs, outputs, parameters):
        train_data = inputs["train data"].value
        responses = inputs["responses"].value
        # print("ANN process_intputs", train_data.shape, responses.shape)
        self.params_ = parameters
        # sample_weights = self.get_sample_weights(responses)  # todo: zrobic to - musimy uwzglednic, ze ANN zwraca floaty!
        sample_weights = None
        log = self.train_and_evaluate(train_data, responses, sample_weights, parameters["CV_k"])
        outputs["model"] = Data(self.ann)
        outputs["log"] = Data(log)

    def train(self, train_data, responses, sample_weights):
        params = self.params_
        term_criteria = (cv.TERM_CRITERIA_MAX_ITER + cv.TERM_CRITERIA_EPS, params["max_iter"], params["epsilon"])
        layers = np.array([train_data.shape[1]] + [int(i) for i in params["layer_sizes"].split()] + [
            responses.shape[1] if len(responses.shape) > 1 else 1])
        flags = 0 + params["scaling"] + params["train_mode"]
        print("Training ANN with", train_data.shape[0], "samples... Layers:", layers)
        if self.ann is None or params["train_mode"] == 0:
            self.ann = cv.ml.ANN_MLP_create()
    
            self.ann.setLayerSizes(layers)
            self.ann.setTrainMethod(params['train_method'])
            self.ann.setBackpropMomentumScale(params['moment_scale'])
            self.ann.setBackpropWeightScale(params['dw_scale'])
            self.ann.setRpropDW0(params['dw0'])
            self.ann.setRpropDWMin(params['dw_min'])
            self.ann.setTermCriteria(term_criteria)
            self.ann.setActivationFunction(params['activate_func'])

            flags -= params['train_mode']

        data = cv.ml.TrainData_create(train_data, cv.ml.ROW_SAMPLE, responses)
        iters = self.ann.train(data, flags)
        print("Done training ANN. Iterations:", iters)

    def predict(self, v):
        ret = self.ann.predict(v.reshape((1, len(v))))[1][0]
        return np.nan_to_num(ret)  # opencv ANN sometimes returns NaNs... ehhh... let's convert them to 0's

    def retrain(self):
        self.ann = None
        self.recalculate(True, False, True)

        # def evaluate_training(self, train_data, responses):
        #     errors = np.zeros(responses.shape[1])
        #     for v, response in zip(train_data, responses):
        #         output = self.ann.predict(v)
        #         errors += abs(response - output)
        #     errors /= train_data.shape[0]
        #     return "Train data result: average errors={}".format(errors)

    def save(self):
        path, _ = QFileDialog.getSaveFileName(self, "Save neural network")
        if not path: return
        self.ann.save(str(path))


register_elements("Machine Learning - OpenCV", [SvmTrain, SvmTrainNoParams, SvmPredict, AnnMlpTrain], 20)



import os
from subprocess import check_call
from tempfile import mkstemp, mkdtemp

import cv2 as cv

from ...core.threaded_element import ThreadedElement
from ...view.elements import *
from ..data import *
from ..connectors import *
from ..errors import *

from . import register_elements_auto, register_elements


cvtypes = {
    "uint8": cv.CV_8U,
    "uint16": cv.CV_16U,
    "int8": cv.CV_8S,
    "int16": cv.CV_16S,
    "float32": cv.CV_32F,
    "float64": cv.CV_64F
}


class NormalElement(FunctionGuiElement, ThreadedElement):
    pass


class InputElement(InputGuiElement, ThreadedElement):
    def get_source(self):
        name = self.__class__.__name__.lower()
        code = None
        outputs = list(self.outputs.keys())
        return name, code, outputs


class SequenceToDataElement(NormalElement):
    """processes its inputs as-they-are, returns only simple Data type"""

    def get_processing_units(self, inputs, parameters):
        outputs = {name: Data() for name in self.outputs}
        units = [ProcessingUnit(self, inputs, parameters, outputs)]
        return units, outputs



class MultiInputOneOutputElement(NormalElement):
    def get_attributes(self):
        return [Input("inputs", multiple=True)], [Output("output")], []

    def get_processing_units(self, inputs, parameters):
        input_list = inputs["inputs"].value
        if len(input_list) == 1 and input_list[0].type() == Data.SEQUENCE:
            ins = {i: d for i, d in enumerate(input_list[0].value)}
        else:
            ins = {i: d for i, d in enumerate(input_list)}
        return NormalElement.get_default_processing_units(self, ins, parameters, ["output"])


class SequenceToSequenceElement(NormalElement):
    num_outputs = 8
    def get_attributes(self):
        return [Input("inputs", multiple=True)], [Output("output")], []

    def get_processing_units(self, inputs, parameters):
        outputs = {"output": Sequence([ImageData() for _ in range(self.num_outputs)])}
        return [ProcessingUnit(self, inputs, parameters, outputs)], outputs


class ProcessElement(NormalElement):
    command = "command_name {args} {inputs} {outputs}"

    def run_command(self, images, output_count, **args):
        in_files = [mkstemp(".bmp", "cvlab_in_")[1] for _ in images]

        if output_count:
            out_files = [mkstemp(".bmp", "cvlab_out_")[1] for _ in range(output_count)]
            out_dir = None
        else:
            out_files = []
            out_dir = mkdtemp(prefix="cvlab_out_")

        try:
            for in_file, image in zip(in_files, images):
                cv.imwrite(in_file, image)

            command = self.command.format(inputs=" ".join(in_files), outputs=" ".join(out_files), output_dir=out_dir, **args)
            command = command.replace("\\", "/")
            print("Executing:", command)
            check_call(command, shell=True)

            results = []

            if out_files:
                results += [cv.imread(out_file) for out_file in out_files]

            if out_dir:
                outputs = os.listdir(out_dir)
                results += [cv.imread(out_dir + "/" + o) for o in outputs]

            if len(results) == 1: return results[0]
            else: return tuple(results)

        finally:
            try:
                for f in in_files + out_files:
                    os.remove(f)
            except Exception as e:
                print("Exception during ProcessElement cleanup:", e)





from .version import __version__


def main(*args, **kwargs):
    import os
    import sys
    import sip
    import numpy as np

    sys._excepthook = sys.excepthook
    def exception_hook(exctype, value, traceback):
        sys._excepthook(exctype, value, traceback)
        sys.exit(-1)
    sys.excepthook = exception_hook

    # todo: it's an ugly workaround for PyQt stylesheets relative paths
    try:
        os.chdir(os.path.dirname(__file__))
    except Exception:
        pass

    np.seterr(all='raise')
    sip.setdestroyonexit(False)

    from PyQt5.QtWidgets import QApplication
    app = QApplication(sys.argv)

    from .view.mainwindow import MainWindow
    main_window = MainWindow(app)
    ret_code = app.exec_()
    return ret_code


if __name__ == '__main__':
    main()



from collections import defaultdict
from threading import Lock, RLock

from .errors import ProcessingError


class Data:
    NONE = 0
    SEQUENCE = 1
    IMAGE = 2

    def __init__(self, value=None, _type=IMAGE):
        self._type = _type
        if _type == Data.SEQUENCE and value is None:
            self._value = []
        else:
            self._value = value
        self.observers = defaultdict(int)
        self.observers_lock = Lock()
        self.lock = RLock()

    def add_observer(self, observer, recursive):
        with self.observers_lock:
            self.observers[observer] += 1
        if recursive and self._type == Data.SEQUENCE:
            with self.lock:
                for v in self._value:
                    v.add_observer(observer, True)

    def remove_observer(self, observer, recursive):
        with self.observers_lock:
            self.observers[observer] -= 1
            if self.observers[observer] <= 0:
                del self.observers[observer]
        if recursive and self._type == Data.SEQUENCE:
            with self.lock:
                for v in self._value:
                    v.remove_observer(observer, True)

    def clear(self):
        with self.lock:
            if self._type == Data.SEQUENCE:
                for d in self._value:
                    d.clear()
            else:
                self.value = None

    def copy(self):
        with self.lock:
            if self._type == self.NONE:
                return EmptyData()
            if self._type == self.SEQUENCE:
                return Sequence([d.copy() for d in self._value])
            elif self._type == self.IMAGE:
                if self._value is None or (hasattr(self._value, "size") and not len(self._value)):
                    pass
                return ImageData(self._value)
            else:
                raise TypeError("Wrong Data.type")

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_value):
        with self.lock:
            if self._value is new_value:
                return
            self._value = new_value
        with self.observers_lock:
            for o in self.observers:
                o(self)

    def assign(self, other):
        assert isinstance(other, Data)
        if not self.is_compatible(other):
            raise ProcessingError("Data.assign: Data not compatible")
        if self._type == Data.SEQUENCE:
            if len(self._value) != len(other._value):
                raise ProcessingError("Data.assign: Sequence not compatible")
            for mine, her in zip(self._value, other._value):
                mine.assign(her)
        else:
            self.value = other.value

    def is_compatible(self, other):
        assert isinstance(other, Data)
        if self._type != other._type: return False
        if self._type == Data.SEQUENCE:
            if len(self._value) != len(other._value): return False
            return all(mine.is_compatible(her) for mine, her in zip(self._value, other._value))
        return True

    def type(self):
        with self.lock:
            if self._value is None or (self._type == Data.IMAGE and hasattr(self.value, "size") and not self._value.size):
                return Data.NONE
            else:
                return self._type

    def sequence_get_value(self, sequence_number):
        with self.lock:
            if self.type() != Data.SEQUENCE:
                return self
            else:
                return self._value[min(sequence_number, len(self._value) - 1)]

    __getitem__ = sequence_get_value

    def __iter__(self):
        with self.lock:
            if self.type() == Data.NONE:
                return iter([])
            elif self.type() == Data.SEQUENCE:
                return iter(self._value)
            else:
                return iter([self])

    def desequence_all(self):
        """Returns a one-dimensional array with all sequence values"""
        with self.lock:
            if self._type == Data.NONE: return [None]
            if self._type == Data.IMAGE: return [self._value]
            if self._type == Data.SEQUENCE:
                t = []
                for d in self._value:
                    t += d.desequence_all()
                return t
            raise TypeError("Wrong data type - cannot desequence")

    def is_complete(self):
        with self.lock:
            t = self.type()
            if t == Data.NONE:
                return False
            elif t == Data.IMAGE:
                return True
            else:
                return all(d.is_complete() for d in self._value)

    def create_placeholder(self):
        if self._type == Data.SEQUENCE:
            return Data([d.create_placeholder() for d in self._value], Data.SEQUENCE)
        else:
            return Data()

    def __str__(self):
        return repr(self)

    def __repr__(self):
        id_ = "0x{:08X}".format(id(self))
        if self._type == Data.NONE: return '<Data [empty] at {}>'.format(id_)
        if self.type() == Data.IMAGE:
            try:
                shape = "?"
                shape = len(self._value)
                shape = self._value.shape
            except Exception:
                pass
            return '<Data [image {}] at {}>'.format(shape, id_)
        if self.type() == Data.NONE: return "<Data [empty image] at {}>".format(id_)
        if self.type() == Data.SEQUENCE:
            s = ""
            images_count = 0
            none_count = 0
            for d in self._value:
                if d.type() == Data.SEQUENCE:
                    s += str(d) + ", "
                if d.type() == Data.IMAGE:
                    images_count += 1
                if d.type() == Data.NONE:
                    none_count += 1
            if images_count:
                s += str(images_count) + " images, "
            if none_count:
                s += str(none_count) + " nones, "
            if len(s) > 0:
                s = s.rstrip(', ')
            s = '<Sequence ' + '[' + s + '] at {}>'.format(id_)
            return s
        raise TypeError("Wrong data type - cannot desequence")

    def __bool__(self):
        return self.type() != Data.NONE

    ready = __bool__

    def __eq__(self, other):
        if not isinstance(other, Data): return False
        with self.lock, other.lock:
            if self.type() != other.type(): return False
            if self.type() == Data.SEQUENCE:
                if len(self._value) != len(other._value): return False
                assert all(isinstance(d, Data) for d in self._value)
                assert all(isinstance(d, Data) for d in other._value)
                return all(a == b for a, b in zip(self._value, other._value))
            else:
                return self._value is other._value
                # todo: if we want logical equality rather than reference equality, we shall use this:
                # elif self.type() == Data.IMAGE:
                # #assert isinstance(self._value, np.ndarray)
                # return self._value is other._value and np.array_equal(self._value, other._value)
                # else:
                #     raise ValueError("Wrong data type")


class EmptyOptionalData(Data):
    def ready(self):
        return True


def Sequence(values=None):
    return Data(values, Data.SEQUENCE)


def EmptyData():
    return Data(None, Data.NONE)


def ImageData(value=None):
    return Data(value, Data.IMAGE)



class DataSet:
    def __init__(self, inputs=None, parameters=None, outputs=None):
        self.inputs = inputs if inputs is not None else {}
        self.parameters = parameters if parameters is not None else {}
        self.outputs = outputs if outputs is not None else {}

    def reset_outputs(self):
        for d in self.outputs.values():
            assert isinstance(d, Data)
            d.clear()

    def __eq__(self, other):
        return self.parameters == other.parameters and self.inputs == other.inputs

    def __str__(self):
        return "DataSet(" + repr(self.inputs) + ", " + repr(self.parameters) + ", " + repr(self.outputs) + ")"

    __repr__ = __str__


class ProcessingUnit(DataSet):
    def __init__(self, element, inputs=None, parameters=None, outputs=None):
        super(ProcessingUnit, self).__init__(inputs, parameters, outputs)
        self.element = element
        self.calculated = False

    def is_being_processed(self):
        return self.element.actual_processing_unit is self

    def ready_to_execute(self):
        return all(d.ready() for d in self.inputs.values())

    def data_changed(self, data):
        if self.calculated:
            self.calculated = False
            self.reset_outputs()
        self.element.recalculate(False, False, self.is_being_processed())

    def connect_observables(self):
        for input in self.inputs.values():
            input.add_observer(self.data_changed, True)

    def disconnect_observables(self):
        for input in self.inputs.values():
            input.remove_observer(self.data_changed, True)

    def __str__(self):
        return "ProcessingUnit(" + repr(self.inputs) + ", " + repr(self.parameters) + ", " + repr(
            self.outputs) + ", calculated=" + repr(self.calculated) + ", ready_to_execute=" + repr(
            self.ready_to_execute()) + ")"

    __repr__ = __str__

    def __hash__(self):
        return hash(id(self))

    def __eq__(self, other):
        return self is other




import re
from datetime import datetime, timedelta

import numpy as np
from PyQt5 import QtCore
from PyQt5.QtCore import pyqtSignal, pyqtSlot, QObject, Qt, QTimer
from PyQt5.QtWidgets import *
from PyQt5.QtGui import *

from ..diagram.element import Element
from .elements import GuiElement
from .mimedata import Mime
from .wires import WiresForeground, NO_FOREGROUND_WIRES, WiresBackground, WireTools


class ScrolledWorkArea(QScrollArea):
    def __init__(self, diagram, style_manager):
        super(ScrolledWorkArea, self).__init__()
        self.setObjectName("ScrolledWorkArea")
        self.workarea = WorkArea(diagram, style_manager)
        self.workarea.setParent(self)
        self.diagram = diagram
        self.setWidget(self.workarea)
        diagram.set_painter(self)
        self.mouse_press_pos = None
        QTimer.singleShot(50, self.scroll_to_absolute_center)

    def load_diagram_from_json(self, ascii_data):
        self.diagram.load_from_json(ascii_data)
        QTimer.singleShot(100, self.scroll_to_upperleft)

    def element_z_index(self, element):
        return self.workarea.children().index(element)

    def mousePressEvent(self, e):
        self.mouse_press_pos = e.pos() + QtCore.QPoint(self.horizontalScrollBar().value(), self.verticalScrollBar().value())
        e.accept()

    def mouseMoveEvent(self, e):
        if e.buttons() & Qt.MiddleButton and self.mouse_press_pos is not None:
            pos = self.mouse_press_pos - e.pos()
            self.horizontalScrollBar().setValue(pos.x())
            self.verticalScrollBar().setValue(pos.y())
            e.accept()

    def wheelEvent(self, event):
        assert isinstance(event, QWheelEvent)
        if event.modifiers() in (Qt.ControlModifier, Qt.MetaModifier):

            origin = (event.x() + self.horizontalScrollBar().value(),
                      event.y() + self.verticalScrollBar().value())

            if event.angleDelta().y() < 0:
                self.workarea.zoom(index=-1, origin=origin)
            else:
                self.workarea.zoom(index=1, origin=origin)
            event.accept()
        else:
            super(ScrolledWorkArea, self).wheelEvent(event)

    def scroll_to_center(self):
        if not self.diagram.elements:
            self.scroll_to_absolute_center()
            return
        pos, n = QtCore.QPoint(0, 0), 0
        for e in self.diagram.elements:
            n += 1
            pos += e.pos() + QtCore.QPoint(e.width()//2, e.height()//2)
        pos //= n
        pos -= QtCore.QPoint(self.width()//2, self.height()//2)
        self.horizontalScrollBar().setValue(pos.x())
        self.verticalScrollBar().setValue(pos.y())

    def scroll_to_upperleft(self):
        if not self.diagram.elements:
            self.scroll_to_absolute_center()
            return
        left = min(e.x() for e in self.diagram.elements) - 20
        top = min(e.y() for e in self.diagram.elements) - 20
        self.horizontalScrollBar().setValue(left)
        self.verticalScrollBar().setValue(top)

    def scroll_to_absolute_center(self):
        left = self.workarea.width()//2 - (self.width() - self.verticalScrollBar().width())//2
        top = self.workarea.height()//2 - (self.height() - self.horizontalScrollBar().height())//2
        self.horizontalScrollBar().setValue(left)
        self.verticalScrollBar().setValue(top)


class UserActions(QObject):
    element_relocated = pyqtSignal(Element)
    cursor_line_started = pyqtSignal()
    cursor_line_moved = pyqtSignal(tuple)
    cursor_line_dropped = pyqtSignal()


class WorkArea(QWidget):
    zoom_levels = [0.25, 0.5, 0.75, 1.0]
    DEFAULT_POSITION_GRID = 20

    def __init__(self, diagram, style_manager):
        super(WorkArea, self).__init__()
        self.setObjectName("WorkArea")
        self.setAcceptDrops(True)
        self.diagram = diagram
        self.user_actions = UserActions()
        self.style_manager = style_manager
        self.wire_tools = WireTools(style_manager)
        self.wires_in_foreground = WiresForeground(self, self.user_actions, self.wire_tools)
        self.wires_in_background = WiresBackground(self, self.user_actions, self.wire_tools)
        self.selection_manager = SelectionManager(self)
        self.connectors_map = {}
        self.diagram.element_added.connect(self.on_element_added)
        self.diagram.element_deleted.connect(self.on_element_deleted)
        self.element_move_start = None
        self.last_auto_scroll_time = datetime.now()
        self.style_manager.style_changed.connect(self.actualize_style)

    @property
    def position_grid(self):
        return int(round(self.DEFAULT_POSITION_GRID * self.diagram.zoom_level))

    def mousePressEvent(self, e):
        if e.button() == Qt.LeftButton:
            self.selection_manager.on_workarea_mouse_left_pressed(e)
            self.wires_in_background.mousePressEvent(e)
        e.ignore()

    def mouseMoveEvent(self, e):
        if e.buttons() & Qt.LeftButton:
            self.selection_manager.on_workarea_mouse_left_move(e)
        if e.buttons() == Qt.MiddleButton:
            e.ignore()

    def mouseReleaseEvent(self, e):
        if e.button() == Qt.LeftButton:
            self.selection_manager.on_workarea_mouse_left_released(e)

    def on_element_moused_pressed(self, element, event):
        self.element_move_start = event.pos()
        if event.button() == QtCore.Qt.LeftButton:
            self.selection_manager.on_element_moused_left_pressed(element, event)
        if event.button() == QtCore.Qt.RightButton:
            self.selection_manager.on_element_right_pressed(element, event)
        self.wires_in_background.unselect_wires()

    def on_element_moused_left_released(self, element, event):
        if (event.pos() == self.element_move_start):
            self.selection_manager.on_element_left_released_inplace(element, event)
        self.element_move_start = None

    def on_element_moused_left_moved(self, element, event):
        if self.element_move_start is not None:
            for e in self.selection_manager.selected_elements:
                pos = e.pos() - self.element_move_start + event.pos()
                x, y = self.nearest_grid_point(pos.x(), pos.y())
                e.move(x, y)
                e.element_relocated.emit(e)
            if datetime.now() - self.last_auto_scroll_time > timedelta(milliseconds=150):
                cursor_pos = element.pos() + event.pos()
                self.parent().parent().ensureVisible(cursor_pos.x(), cursor_pos.y(), 50, 50)  # pozwala na scrollowanie
                self.last_auto_scroll_time = datetime.now()

    @pyqtSlot(Element, tuple)
    def on_element_added(self, element, position):
        element.setParent(self)
        element.move(position[0], position[1])
        element.setVisible(True)
        element.setFocus()
        element.set_workarea(self)
        element.element_relocated.connect(self.user_actions.element_relocated)
        self.adjustSize()
        if not NO_FOREGROUND_WIRES:
            self.wires_in_foreground.raise_()
        self.connectors_map.update(element.input_connectors)
        self.connectors_map.update(element.output_connectors)

    @pyqtSlot(Element)
    def on_element_deleted(self, element):
        for connector in list(element.outputs.values()) + list(element.inputs.values()):
            del self.connectors_map[connector]
            # Todo: potential memory leak - see: http://stackoverflow.com/questions/5899826/pyqt-how-to-remove-a-widget
        element.setParent(None)
        element.deleteLater()

    def dragEnterEvent(self, e):
        e.accept()

    def dropEvent(self, e):
        mime = e.mimeData().text()
        if mime == Mime.NEW_ELEMENT and e.source().last_spawned_element != 0:
            new_elem = e.source().last_spawned_element
            position = (e.pos().x() - new_elem.width()//2,
                        e.pos().y() - new_elem.height()//2)
            self.diagram.add_element(new_elem, position)
        elif mime in [Mime.INCOMING_CONNECTION, Mime.OUTGOING_CONNECTION]:
            self.user_actions.cursor_line_dropped.emit()
        e.accept()

    def dragMoveEvent(self, e):
        mime = e.mimeData().text()
        # todo: forcing update here causes GUI lags REGARDLESS of the Lines painEvent() (method can be even empty)
        # - the problem occurs at least on OS X
        # - possible solutions is updating parts of the Lines widget instead of all of it, by detecting possibly 3
        #   rectangles for each line that is being affected by the drag:
        #   http://stackoverflow.com/questions/18043492/qt-custom-widget-update-big-overhead
        # - some improvement can be made to paintEvent() - using QImage to draw lines and then dump it to painter,
        #   which will use software renderer OR/AND QGLWidget:
        #   http://stackoverflow.com/questions/6089642/qpainter-painting-alternatives-performance-sucks-on-mac
        if mime == Mime.INCOMING_CONNECTION:
            points = (e.source(), e.pos())
            self.user_actions.cursor_line_moved.emit(points)
        elif mime == Mime.OUTGOING_CONNECTION:
            points = (e.pos(), e.source())
            self.user_actions.cursor_line_moved.emit(points)

    def sizeHint(self):
        default_size = 30000
        layout = self.parent().parent()
        united = QtCore.QRect(QtCore.QPoint(0, 0), layout.maximumViewportSize())
        for e in self.children():
            if isinstance(e, Element):
                rect = e.geometry()
                rect.setWidth(rect.width() + 50)
                rect.setHeight(rect.height() + 50)
                united = united.united(rect)
        return QtCore.QSize(max(default_size, united.right()), max(default_size, united.bottom()))

    # def paintEvent(self, e):
    #     # todo: it's very inefficient (paint event occurs very often)
    #     self.adjustSize()

    def resizeEvent(self, e):
        self.wires_in_foreground.setGeometry(self.rect())
        self.wires_in_background.setGeometry(self.rect())

    def zoom(self, level=None, index=None, origin=None):
        assert (level is None and index is not None) or (level is not None and index is None)

        if index is not None:
            new_level = None
            if index > 0:
                for i in range(len(self.zoom_levels)):
                    if self.zoom_levels[i] > self.diagram.zoom_level:
                        new_level = i
                        break
            else:
                for i in reversed(range(len(self.zoom_levels))):
                    if self.zoom_levels[i] < self.diagram.zoom_level:
                        new_level = i
                        break
            if new_level is None:
                return

            level = self.zoom_levels[new_level]

        factor = level / self.diagram.zoom_level
        self.diagram.zoom_level = level

        origin = origin or (min(e.pos().x() for e in self.diagram.elements),
                            min(e.pos().y() for e in self.diagram.elements))

        for e in self.diagram.elements:
            assert isinstance(e, GuiElement)
            e.zoom(factor, origin)

        self.actualize_style()

    def actualize_style(self):
        def sub(match):
            value = int(match.group(1))
            if value == 0:
                return "0px"
            value = int(round(value * self.diagram.zoom_level))
            value = max(value, 1)
            return str(value) + "px"

        style = self.style_manager.main_window.styleSheet()
        style = re.sub(r"(\d+)\s*px", sub, style)
        self.setStyleSheet(style)

    def nearest_grid_point(self, x, y):
        return int(round(float(x)/self.position_grid) * self.position_grid), int(round(float(y)/self.position_grid) * self.position_grid)


class SelectionManager:

    def __init__(self, workarea):
        self.workarea = workarea
        self.rubberBand = QRubberBand(QRubberBand.Rectangle, self.workarea)
        self.selection_origin = None
        self.selected_elements = []
        self.last_selection = []
        self.workarea.diagram.element_deleted.connect(self.on_element_deleted)

    def selected_count(self):
        return len(self.selected_elements)

    def on_element_moused_left_pressed(self, element, event):
        ctrl_pressed = self.is_control_pressed()
        if not element.selected:
            if not ctrl_pressed:
                self.clear_selection()
            self.select_element(element)
        else:
            if ctrl_pressed:
                self.unselect_element(element)
            else:
                self.select_element(element)

    def on_element_right_pressed(self, element, event):
        if not element.selected:
            self.clear_selection()
            self.select_element(element)

    def on_element_left_released_inplace(self, element, event):
        ctrl_pressed = self.is_control_pressed()
        if not ctrl_pressed:
            self.clear_selection()
            self.select_element(element)

    def on_workarea_mouse_left_pressed(self, event):
        if not self.is_control_pressed():
            self.clear_selection()
        self.selection_origin = QtCore.QPoint(event.pos())
        self.rubberBand.setGeometry(QtCore.QRect(self.selection_origin, QtCore.QSize()))
        self.rubberBand.show()
        self.last_selection = self.selected_elements[:]

    def on_workarea_mouse_left_move(self, event):
        if self.selection_origin is not None and not self.selection_origin.isNull():
            rect = QtCore.QRect(self.selection_origin, event.pos()).normalized()
            # FIXME: Hide and show prevent a strange effect of QRubberBand of shaking the widgets beneath, which occur
            # when making selection in a direction different the right-bottom, but only after placing an element from
            # the palette - on a freshly loaded diagram everything is fine
            self.rubberBand.hide()
            self.rubberBand.setGeometry(rect)
            self.rubberBand.show()
            for e in self.workarea.diagram.elements:
                ctrl_pressed = self.is_control_pressed()
                if self.is_element_in_rect(e, rect):
                    if ctrl_pressed and e in self.last_selection:
                        self.unselect_element(e)
                    else:
                        self.select_element(e)
                else:
                    if not ctrl_pressed:
                        self.unselect_element(e)

    def on_workarea_mouse_left_released(self, event):
        self.rubberBand.hide()
        self.last_selection[:] = []

    def delete_selected(self):
        while len(self.selected_elements) != 0:
            self.workarea.diagram.delete_element(self.selected_elements[0])

    def select_element(self, element):
        if not element.selected:
            self.selected_elements.append(element)
            element.set_selected(True)

    def unselect_element(self, element):
        if element.selected:
            if element in self.selected_elements:
                self.selected_elements.remove(element)
            element.set_selected(False)

    def clear_selection(self):
        for e in self.selected_elements:
            e.set_selected(False)
        self.selected_elements = []

    def is_element_in_rect(self, element, rect):
        return rect.contains(element.frameGeometry())

    def is_control_pressed(self):
        return (int(QApplication.keyboardModifiers()) & QtCore.Qt.ControlModifier) != 0

    def on_element_deleted(self, element):
        if element in self.selected_elements:
            self.selected_elements.remove(element)
        if element in self.last_selection:
            self.last_selection.remove(element)



import inspect
import re

from .hooks import *
from .exceptions import *


TEST_QT = False


class CoreElement(Element):
    """Base class for all diagram elements"""

    def __init__(self):
        super(CoreElement, self).__init__()
        for o in self.outputs.values():
            o.hook = OutputHook(o) if not TEST_QT else OutputQtHook(o)
        for i in self.inputs.values():
            i.hook = InputHook(i) if not TEST_QT else InputQtHook(i)
        self.state = self.STATE_READY
        self.data = None
        self.units = []
        self.parameters_changed = True
        self.structure_changed = True
        self.actual_processing_unit = None
        self.delayed_recalculate = False
        self.is_recalculating = False
        self.prepare_empty_data()

    def recalculate(self, refresh_parameters, refresh_structure, force_break, force_units_recalc=False):
        while True:
            try:
                self.parameters_changed |= refresh_parameters
                self.structure_changed |= refresh_structure
                if self.is_recalculating:
                    return
                else:
                    self.delayed_recalculate = False
                self.is_recalculating = True
                self.set_state(self.STATE_BUSY)
                for unit in self.units:
                    unit.calculated = False
                self.process()
                self.set_state(self.STATE_READY)
            except (InterruptException, ProcessingBreak):
                pass
            except Exception as e:
                self.set_state(self.STATE_ERROR, e)  # todo: core2: uncomment this!!!
            self.actual_processing_unit = None
            if not self.delayed_recalculate:
                break
        self.is_recalculating = False

    def may_interrupt(self):
        """Informs the program that element may be interrupted here"""
        return self.delayed_recalculate

    def process_channels(self, inputs, outputs, parameters):
        """Does processing of input channels and returns the outputs"""
        print("process channels")

    def process_inputs(self, inputs, outputs, parameters):
        """Does processing of inputs and returns the outputs"""

        ins = defaultdict(dict)  # [channel number][name] -> input channel image
        for iname, idata in inputs.items():
            for chnum, ich in enumerate(cv.split(idata.value)):
                ins[chnum][iname] = Data(ich)
        outs = defaultdict(dict)  # [output name][channel number] -> output channel image
        for chnum, ich in ins.items():
            o = {}
            self.may_interrupt()
            self.process_channels(ich, o, parameters)
            self.may_interrupt()
            for outname, outch in o.items():
                outs[outname][chnum] = outch
        for outname, channels in outs.items():
            if not all(channels.values()):
                outputs[outname] = Data()
            else:
                outputs[outname] = Data(cv.merge([i.value for i in channels.values()]))

    def get_processing_units(self, inputs, parameters):
        """Returns processing units and output placeholders for the element"""
        return self.get_default_processing_units(inputs, parameters, self.outputs.keys())

    def get_default_processing_units(self, inputs, parameters, output_ids):
        sequences = 0

        for input_ in inputs.values():
            if input_.type() == Data.SEQUENCE:
                sequences = max(sequences, len(input_.value))
            else:
                sequences = max(sequences, 0)

        if sequences == 0:
            outputs = {name: Data() for name in output_ids}
            return [ProcessingUnit(self, inputs, parameters, outputs)], outputs

        units = []
        outputs = {name: Sequence() for name in output_ids}

        for seq_number in range(0, sequences):
            seq_inputs = {input_name: input_.sequence_get_value(seq_number) for input_name, input_ in inputs.items()}
            seq_units, seq_outputs = self.get_default_processing_units(seq_inputs, parameters, output_ids)
            units += seq_units
            for output_name, output_data in seq_outputs.items():
                outputs[output_name].value.append(output_data)

        return units, outputs

    def clear_outputs(self):
        for o in self.data.outputs.values():
            o.clear()

    def prepare_parameters(self):
        self.parameters_changed = False
        for name, parameter in self.parameters.items():
            self.data.parameters[name] = parameter.get()
        self.clear_outputs()
        for unit in self.units:
            unit.calculated = False

    def prepare_empty_data(self):
        self.data = DataSet()
        for unit in self.units:
            unit.disconnect_observables()
        self.units = []

    def prepare_structure(self):
        self.structure_changed = False
        self.prepare_empty_data()
        self.prepare_parameters()
        for name, input_ in self.inputs.items():
            self.data.inputs[name] = input_.get()
        self.units, self.data.outputs = self.get_processing_units(self.data.inputs, self.data.parameters)
        for name, data in self.data.outputs.items():
            self.outputs[name].put(data)
        for unit in self.units:
            unit.connect_observables()

    def prepare_data(self):
        if self.structure_changed:
            self.prepare_structure()
        elif self.parameters_changed:
            self.prepare_parameters()

    def process_units(self):
        for unit in self.units:
            assert isinstance(unit, ProcessingUnit)
            if unit.calculated:
                continue
            if not unit.ready_to_execute():
                continue
            self.actual_processing_unit = unit
            self.may_interrupt()
            inputs = {n: d.copy() for n, d in unit.inputs.items()}
            self.may_interrupt()
            outputs = {}
            self.process_inputs(inputs, outputs, unit.parameters)
            self.may_interrupt()

            # TODO: This is a workaround. Elements should never remove objects from 'outputs'
            # We should really modify Elements to no stop doing that
            # Also, this is probably wrong if there is more than one ProcessingUnit!
            for output_name in outputs:
                unit.outputs[output_name].assign(outputs[output_name])

            unit.calculated = True
            self.actual_processing_unit = None
            self.may_interrupt()

    def process(self):
        self.prepare_data()
        self.process_units()
        self.may_interrupt()

    def delete(self):
        self.prepare_empty_data()  # disconnects data connections
        for o in self.outputs.values():
            o.disconnect_all()
        for i in self.inputs.values():
            i.disconnect_all()
            i.hook.delete()

    def get_source(self):
        """Returns FUNCTION_NAME and SOURCE of the element in format (without leading whitespace!)
            def FUNCTION_NAME(inputs, outputs, parameters):
                # do processing: inputs, parameters -> outputs
        """
        name = self.__class__.__name__.lower()
        if self.process_inputs != CoreElement.process_inputs:
            src = inspect.getsource(self.process_inputs).replace("def process_inputs(self,", "def " + name + "(")
        else:
            src = inspect.getsource(self.process_channels).replace("def process_channels(self,", "def " + name + "(")
        src = src.replace("self.may_interrupt()", "")
        margin = re.search(r"^(\s+)def ", src)
        if margin: src = re.sub("^" + margin.group(1), "", src, flags=re.MULTILINE)
        src = src.replace("\t", "    ")
        return name, src, []




class ProcessingTimeInfo:

    def __init__(self, start, end, units_count, previous_time_infos):
        milis = int(1000 * (end - start))
        units = units_count if units_count != 0 else 1
        self.work_time = milis
        self.work_time_per_unit = int(float(milis) / units)
        self.total_work_time = milis
        if len(previous_time_infos):
            previous_work_time = 0
            for time in previous_time_infos:
                if time is not None:
                    previous_work_time = max(previous_work_time, time.total_work_time)
                    self.total_work_time += previous_work_time
        self.total_work_time_per_unit = int(round(self.total_work_time / float(units)))




from cvlab.diagram.elements.base import *


class DSLR(ProcessElement):
    name = "DSLR Camera"
    comment = "DSLR Camera photo capture"

    lock = Lock()
    command = "cd /tmp; killall gvfsd-gphoto2; killall gvfs-gphoto2-volume-monitor; gphoto2 --capture-image-and-download --force-overwrite --filename={outputs}"

    def get_attributes(self):
        return [], \
               [Output("output")], \
               [ButtonParameter("capture", self.capture)]

    def get_processing_units(self, inputs, parameters):
        outputs = {"output": ImageData()}
        units = [ProcessingUnit(self, inputs, parameters, outputs)]
        return units, outputs

    def process_inputs(self, inputs, outputs, parameters):
        with self.lock:
            image = self.run_command([], 1)
        outputs["output"] = ImageData(image)

    def capture(self):
        self.recalculate(True, True, True)

register_elements_auto(__name__, locals(), "Camera", 10)




 # ############################################################################
 #
 # Copyright (c) Microsoft Corporation. 
 #
 # Available under the Microsoft PyKinect 1.0 Alpha license.  See LICENSE.txt
 # for more information.
 #
 # ###########################################################################/

"""defines the core data structures used for communicating w/ the Kinect APIs"""




import ctypes
from ctypes import Array
from ..nui import _NUIDLL


NUI_SKELETON_COUNT = 6

class _EnumerationType(type(ctypes.c_int)):
    """metaclass for an enumeration like type for ctypes"""

    def __new__(metacls, name, bases, dict):
        cls = type(ctypes.c_int).__new__(metacls, name, bases, dict)
        for key, value in list(cls.__dict__.items()):                        
            if key.startswith('_') and key.endswith('_'): continue
            
            setattr(cls, key, cls(key, value))
        
        return cls


class _Enumeration(with_metaclass(_EnumerationType, ctypes.c_int)):
    """base class for enumerations"""
    def __init__(self, name, value):
        self.name = name        
        ctypes.c_int.__init__(self, value)
      
    def __hash__(self):
        return self.value

    def __int__(self):
        return self.value
    
    def __index__(self):
        return self.value
      
    def __repr__(self):
        if hasattr(self, 'name'):
            return "<%s.%s (%r)>" % (self.__class__.__name__, self.name, self.value)

        name = '??'
        for x in type(self).__dict__:
            if x.startswith('_') and x.endswith('_'): continue

            if getattr(self, x, None).value == self.value:
                name = x
                break

        return "<%s.%s (%r)>" % (self.__class__.__name__, name, self.value)

    def __eq__(self, other):
        if type(self) is not type(other):
            return self.value == other
            
        return self.value == other.value

    def __ne__(self, other):
        if type(self) is not type(other):
            return self.value != other

        return self.value != other.value


class Vector(ctypes.Structure):
    """Represents vector data."""
    _fields_ = [('x', ctypes.c_float),
            ('y', ctypes.c_float),
            ('z', ctypes.c_float),
            ('w', ctypes.c_float)
            ]

    def __init__(self, x = 0.0, y = 0.0, z = 0.0, w = 0.0):
        self.x = x
        self.y = y
        self.z = z
        self.w = w

    def __eq__(self, other):
        return (self.x == other.x and
               self.y == other.y and
               self.z == other.z and
               self.w == other.w)
 
    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '<x=%r, y=%r, z=%r, w=%r>' % (self.x, self.y, self.z, self.w)


class Matrix4(Array):
    """4x4 matrix.  Can be accessed using matrix[0,0] ... matrix[3,3] or can be accessed using
       matrix.M11 ... matrix.M44 for similarity to .NET and the C data structures.  matrix[0,1] is
       the same as matrix.M12.
    
        Used to provide bone rotation information.   
    """

    _length_ = 16
    _type_ = ctypes.c_float

    def __getitem__(self, index):
        return Array.__getitem__(self, index[1] + index[0] * 4)

    def __setitem__(self, index, value):
        return Array.__setitem__(self, index[1] + index[0] * 4, value)

    def get_M11(self): return Array.__getitem__(0)
    def set_M11(self, value): Array.__setitem__(0, value)
    M11 = property(get_M11, set_M11)

    def get_M12(self): return Array.__getitem__(1)
    def set_M12(self, value): Array.__setitem__(1, value)
    M12 = property(get_M12, set_M12)

    def get_M13(self): return Array.__getitem__(2)
    def set_M13(self, value): Array.__setitem__(2, value)
    M13 = property(get_M13, set_M13)

    def get_M14(self): return Array.__getitem__(3)
    def set_M14(self, value): Array.__setitem__(3, value)
    M14 = property(get_M14, set_M14)

    def get_M21(self): return Array.__getitem__(4)
    def set_M21(self, value): Array.__setitem__(4, value)
    M21 = property(get_M21, set_M21)

    def get_M22(self): return Array.__getitem__(5)
    def set_M22(self, value): Array.__setitem__(5, value)
    M22 = property(get_M22, set_M22)

    def get_M23(self): return Array.__getitem__(6)
    def set_M23(self, value): Array.__setitem__(6, value)
    M23 = property(get_M23, set_M23)

    def get_M24(self): return Array.__getitem__(7)
    def set_M24(self, value): Array.__setitem__(7, value)
    M24 = property(get_M24, set_M24)

    def get_M31(self): return Array.__getitem__(8)
    def set_M31(self, value): Array.__setitem__(8, value)
    M31 = property(get_M31, set_M31)

    def get_M32(self): return Array.__getitem__(9)
    def set_M32(self, value): Array.__setitem__(9, value)
    M32 = property(get_M32, set_M32)

    def get_M33(self): return Array.__getitem__(10)
    def set_M33(self, value): Array.__setitem__(10, value)
    M33 = property(get_M33, set_M33)

    def get_M34(self): return Array.__getitem__(11)
    def set_M34(self, value): Array.__setitem__(11, value)
    M34 = property(get_M34, set_M34)

    def get_M41(self): return Array.__getitem__(12)
    def set_M41(self, value): Array.__setitem__(12, value)
    M41 = property(get_M41, set_M41)

    def get_M42(self): return Array.__getitem__(13)
    def set_M42(self, value): Array.__setitem__(13, value)
    M42 = property(get_M42, set_M42)

    def get_M43(self): return Array.__getitem__(14)
    def set_M43(self, value): Array.__setitem__(14, value)
    M43 = property(get_M43, set_M43)

    def get_M44(self): return Array.__getitem__(15)
    def set_M44(self, value): Array.__setitem__(15, value)
    M44 = property(get_M44, set_M44)


class _NuiLockedRect(ctypes.Structure):
    _fields_ = [('pitch', ctypes.c_int32), 
                ('size', ctypes.c_int32),
                ('bits', ctypes.c_voidp)]

     
class _NuiSurfaceDesc(ctypes.Structure):
    _fields_ = [('width', ctypes.c_uint32),
                ('height', ctypes.c_uint32)
               ]

class PlanarImage(ctypes.c_voidp):
    """Represents a video image."""
    _BufferLen = ctypes.WINFUNCTYPE(ctypes.HRESULT, ctypes.c_int32)(3, 'BufferLen')
    _Pitch = ctypes.WINFUNCTYPE(ctypes.HRESULT, ctypes.c_int32)(4, 'Pitch')
    _LockRect = ctypes.WINFUNCTYPE(ctypes.HRESULT, ctypes.c_uint, ctypes.POINTER(_NuiLockedRect), ctypes.c_voidp, ctypes.c_uint32)(5, '_LockRect')
    _GetLevelDesc = ctypes.WINFUNCTYPE(ctypes.HRESULT, ctypes.c_uint32, ctypes.POINTER(_NuiSurfaceDesc))(6, '_GetLevelDesc')
    _UnlockRect = ctypes.WINFUNCTYPE(ctypes.HRESULT, ctypes.c_uint32)(7, '_UnlockRect')

    @property
    def width(self):
        desc = _NuiSurfaceDesc()
        PlanarImage._GetLevelDesc(self, 0, ctypes.byref(desc))
        return desc.width
    
    @property
    def height(self):
        desc = _NuiSurfaceDesc()
        PlanarImage._GetLevelDesc(self, 0, ctypes.byref(desc))
        return desc.height.value
        
    @property
    def bytes_per_pixel(self):
        return self.pitch / self.width

    @property
    def bits(self):
        buffer = (ctypes.c_byte * self.buffer_length)()
        self.copy_bits(buffer)
        return buffer

    def copy_bits(self, dest):
        """copies the bits of the image to the provided destination address"""
        desc = _NuiSurfaceDesc()
        
        PlanarImage._GetLevelDesc(self, 0, ctypes.byref(desc))

        rect = _NuiLockedRect()
        PlanarImage._LockRect(self, 0, ctypes.byref(rect), None, 0)
        ctypes.memmove(dest, rect.bits, desc.height * rect.pitch)
        PlanarImage._UnlockRect(self, 0)

    @property
    def buffer_length(self):
        return self.width * self.height * self.bytes_per_pixel

    @property
    def pitch(self):
        rect = _NuiLockedRect()
        PlanarImage._LockRect(self, 0, ctypes.byref(rect), None, 0)
        res = rect.pitch
        PlanarImage._UnlockRect(self, 0)
        return res


class ImageType(_Enumeration):
    """Specifies an image type. """
    depth_and_player_index = DepthAndPlayerIndex = 0                 # USHORT
    color = Color = 1                               # RGB32 data
    color_yuv = ColorYuv = 2                            # YUY2 stream from camera h/w, but converted to RGB32 before user getting it.
    color_yuv_raw = ColorYuvRaw = 3                         # YUY2 stream from camera h/w.
    depth = Depth = 4                               # USHORT


class ImageResolution(_Enumeration):
    """Specifies image resolution."""
    invalid = Invalid = -1
    resolution_80x60 = Resolution80x60 = 0 
    resolution_320x240 = Resolution320x240 = 1
    resolution_640x480 = Resolution640x480 = 2
    resolution_1280x1024 = Resolution1280x1024 = 3                      # for hires color only


class SkeletonTracking(_Enumeration):
    suppress_no_frame_data       = 0x00000001   # Prevents NuiSkeletonGetNextFrame from returning E_NUI_FRAME_NO_DATA errors. Instead, calls to NuiSkeletonGetNextFrame block until data is available or the timeout period passes.
    title_sets_tracked_skeletons = 0x00000002   # Disables the default player selection mode and enables the title to manage which players have tracked skeletons.
    enable_seated_support        = 0x00000004   # Uses seated skeleton tracking mode. The 10 lower-body joints of each skeleton will not be tracked.
    enable_in_near_range         = 0x00000008

class ImageDigitalZoom(_Enumeration):
    """Specifies the zoom factor."""

    zoom_1x = Zoom1x = 0  # A zoom factor of 1.0.
    zoom_2x = Zoom2x = 1  # A zoom factor of 2.0.


class ImageViewArea(ctypes.Structure):
    """Specifies the image view area. """
    _fields_ = [('Zoom', ctypes.c_int),     # An ImageDigitalZoom value that specifies the zoom factor. 
                ('CenterX', ctypes.c_long), # The horizontal offset from center, for panning. 
                ('CenterY', ctypes.c_long)  # The vertical offset from center, for panning. 
               ]

    def get_zoom(self):
        return self.Zoom

    def set_zoom(self, value):
        self.Zoom = value

    zoom = property(get_zoom, set_zoom)

    def get_center_x(self):
        return self.CenterX

    def set_center_x(self, value):
        self.CenterX = value

    def get_center_y(self):
        return self.CenterY

    center_x = property(get_center_x, set_center_x)

    def set_center_y(self, value):
        self.CenterY = value

    center_y = property(get_center_y, set_center_y)


class ImageFrame(ctypes.Structure):
    _fields_ = [('timestamp', ctypes.c_longlong),               # The timestamp (in milliseconds) of the most recent frame. The clock starts when you call Initialize.
               ('frame_number', ctypes.c_uint32),              # Returns the frame number
               ('type', ImageType),                         # An ImageType value that specifies the image type.
               ('resolution', ImageResolution),                # An ImageResolution value that specifies the image resolution.
               ('image', PlanarImage),                         # A PlanarImage object that represents the image.
               ('flags', ctypes.c_uint32),                     # flags, not used
               ('view_area', ImageViewArea),                   # An ImageViewArea value that specifies the view area.
              ]


class JointId(_Enumeration):
    """Specifies the various skeleton joints. """
    hip_center = HipCenter = 0
    spine = Spine = 1
    shoulder_center = ShoulderCenter = 2
    head = Head = 3
    shoulder_left = ShoulderLeft = 4
    elbow_left = ElbowLeft = 5
    wrist_left = WristLeft = 6
    hand_left = HandLeft = 7
    shoulder_right = ShoulderRight = 8
    elbow_right = ElbowRight = 9 
    wrist_right = WristRight = 10
    hand_right = HandRight = 11
    hip_left = HipLeft = 12
    knee_left = KneeLeft = 13
    ankle_left = AnkleLeft = 14
    foot_left = FootLeft = 15 
    hip_right = HipRight = 16
    knee_right = KneeRight = 17
    ankle_right = AnkleRight = 18
    foot_right = FootRight = 19
    count = Count = 20

class SkeletonBoneRotation(ctypes.Structure):
    _fields_ = [('rotation_matrix', Matrix4), 
                ('rotation_quaternion', Vector)]

    def __repr__(self):
        return '<SkeletonBoneRotation(%r, %r)>' % (self.rotation_matrix, self.rotation_quaternion)

class SkeletonBoneOrientation(ctypes.Structure):
    _fields_ = [('end_joint', JointId), 
                ('start_joint', JointId),
                ('hierarchical_rotation', SkeletonBoneRotation),
                ('absolute_rotation', SkeletonBoneRotation),
                ]

    def __repr__(self):
        return '<SkeletonBoneOrientation(%r, %r, %r, %r)>' % (self.end_joint, self.start_joint, self.hierarchical_rotation, self.absolute_rotation)


class JointTrackingState(_Enumeration):
    """Specifies the joint tracking state. """
    not_tracked = NOT_TRACKED = 0
    inferred = INFERRED = 1
    tracked = TRACKED = 2


class SkeletonTrackingState(_Enumeration):
    """Specifies a skeleton's tracking state."""
    not_tracked = NOT_TRACKED = 0
    position_only = POSITION_ONLY = 1
    tracked = TRACKED = 2

class SkeletonFrameQuality(_Enumeration):
    """Specifies skeleton frame quality. """
    camera_motion = CameraMotion = 0x01
    extrapolated_floor = ExtrapolatedFloor 	 = 0x02
    upper_body_skeleton = UpperBodySkeleton = 0x04
    seated_support_enabled = 0x08


class SkeletonQuality(_Enumeration):
    """Specifies how much of the skeleton is visible. """
    clipped_right = ClippedRight  = 0x00000001
    clipped_left = ClippedLeft   = 0x00000002
    clipped_top = ClippedTop    = 0x00000004
    clipped_bottom = ClippedBottom = 0x00000008

NUI_SKELETON_POSITION_COUNT = JointId.Count.value

class SkeletonData(ctypes.Structure):
    """Contains data that characterizes a skeleton."""
    _fields_ = [('eTrackingState', SkeletonTrackingState),
                ('dwTrackingID', ctypes.c_uint32),
                ('dwEnrollmentIndex', ctypes.c_uint32),
                ('dwUserIndex', ctypes.c_uint32),
                ('Position', Vector),
                ('SkeletonPositions', ctypes.ARRAY(Vector, NUI_SKELETON_POSITION_COUNT)),
                ('eSkeletonPositionTrackingState', ctypes.ARRAY(JointTrackingState, NUI_SKELETON_POSITION_COUNT)),
                ('Quality', SkeletonQuality),
                ]

    def get_tracking_state(self):
        return self.eTrackingState

    def set_tracking_state(self, value):
        self.eTrackingState = value

    tracking_state = property(get_tracking_state, set_tracking_state)

    def get_tracking_id(self):
        return self.dwTrackingID

    def set_tracking_id(self, value):
        self.dwTrackingID = value

    tracking_id = property(get_tracking_id, set_tracking_id)

    def get_enrollment_index(self):
        return self.dwEnrollmentIndex

    def set_enrollment_index(self, value):
        self.dwEnrollmentIndex = value

    enrollment_index = property(get_enrollment_index, set_enrollment_index)

    def get_user_index(self):
        return self.dwUserIndex

    def set_user_index(self, value):
        self.dwUserIndex = value

    user_index = property(get_user_index, set_user_index)

    def get_position(self):
        return self.Position

    def set_position(self, value):
        self.Position = value

    position = property(get_position, set_position)

    def get_skeleton_positions(self):
        return self.SkeletonPositions

    def set_skeleton_positions(self, value):
        self.SkeletonPositions = value

    skeleton_positions = property(get_skeleton_positions, set_skeleton_positions)

    def get_skeleton_position_tracking_states(self):
        return self.eSkeletonPositionTrackingState

    def set_skeleton_position_tracking_states(self, value):
        self.eSkeletonPositionTrackingState = value

    skeleton_position_tracking_states = property(get_skeleton_position_tracking_states, 
                                                set_skeleton_position_tracking_states)
        
    def get_skeleton_quality(self):
        return self.Quality

    def set_skeleton_quality(self, value):
        self.Quality = value

    skeleton_quality = property(get_skeleton_quality, set_skeleton_quality)

    def calculate_bone_orientations(self):
        """Calculate bone orientations for a skeleton.

        The function calculates hierarchical and absolute joint angles for the skeleton, which can
        be used in animating an avatar (Avateering). The HipCenter joint is the root of the hierarchy,
        and describes an absolute rotation in the right-hand camera coordinate system. All other
        joints describe rotations relative to their parent joint orientation. The angles are returned
        in the same order as the joints are defined.

        Returns a sequence of SkeletonBoneOrientation objects."""
        arr = (SkeletonBoneOrientation*JointId.Count)()
        _NuiSkeletonCalculateBoneOrientations(self, arr)
        return tuple(arr)

    def __repr__(self):
        return '<Tracking: %r, ID: %r, Position: %r>' % (self.eTrackingState, 
                                                            self.dwTrackingID, 
                                                            self.Position)
    def __eq__(self, other):
        if (self.tracking_state == other.tracking_state and
                self.tracking_id == other.tracking_id and
                self.enrollment_index == other.enrollment_index and
                self.user_index == other.user_index and
                self.position == other.position and
                self.skeleton_quality == other.skeleton_quality):

            for i in range(len(self.skeleton_positions)):
                if (self.skeleton_positions[i] != other.skeleton_positions[i] or 
                    self.skeleton_position_tracking_states[i] != other.skeleton_position_tracking_states[i]):
                    return False                 

            return True

        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __bool__(self):
        return self.tracking_state != SkeletonTrackingState.not_tracked

_NuiSkeletonCalculateBoneOrientations = _NUIDLL.NuiSkeletonCalculateBoneOrientations
_NuiSkeletonCalculateBoneOrientations.argtypes = [ctypes.POINTER(SkeletonData), ctypes.POINTER(SkeletonBoneOrientation)]
_NuiSkeletonCalculateBoneOrientations.restype = ctypes.HRESULT


class SkeletonFrame(ctypes.Structure):
    _pack_ = 16
    _fields_ = [('liTimeStamp', ctypes.c_longlong),
                ('dwFrameNumber', ctypes.c_uint32),
                ('Quality', SkeletonFrameQuality),
                ('vFloorClipPlane', Vector),
                ('vNormalToGravity', Vector),
                ('SkeletonData', ctypes.ARRAY(SkeletonData, NUI_SKELETON_COUNT)),
                ]

    def get_timestamp(self):
        return self.liTimeStamp
    
    def set_timestamp(self, value):
        self.liTimeStamp = value
    
    timestamp = property(get_timestamp, set_timestamp)

    def get_frame_number(self):
        return self.dwFrameNumber
    
    def set_frame_number(self, value):
        self.dwFrameNumber = value
    
    frame_number = property(get_frame_number, set_frame_number)

    def get_quality(self):
        return self.Quality
    
    def set_quality(self, value):
        self.Quality = value
    
    quality = property(get_quality, set_quality)

    def get_floor_clip_plane(self):
        return self.vFloorClipPlane
    
    def set_floor_clip_plane(self, value):
        self.vFloorClipPlane = value
    
    floor_clip_plane = property(get_floor_clip_plane, set_floor_clip_plane)

    def get_normal_to_gravity(self):
        return self.vNormalToGravity
    
    def set_normal_to_gravity(self, value):
        self.vNormalToGravity = value
    
    normal_to_gravity = property(get_normal_to_gravity, set_normal_to_gravity)

    def get_skeleton_data(self):
        return self.SkeletonData
    
    def set_skeleton_data(self, value):
        self.SkeletonData = value
    
    skeleton_data = property(get_skeleton_data, set_skeleton_data)


class TransformSmoothParameters(ctypes.Structure):
    """Contains transform smoothing parameters. """
    _fields_ = [('fSmoothing', ctypes.c_float),
                ('fCorrection', ctypes.c_float),
                ('fPrediction', ctypes.c_float),
                ('fJitterRadius', ctypes.c_float),
                ('fMaxDeviationRadius', ctypes.c_float)
                ]

    def get_smoothing(self):
        return self.fSmoothing
    
    def set_smoothing(self, value):
        self.fSmoothing = value
    
    smoothing = property(get_smoothing, set_smoothing)

    def get_correction(self):
        return self.fCorrection
    
    def set_correction(self, value):
        self.fCorrection = value
    
    correction = property(get_correction, set_correction)

    def get_prediction(self):
        return self.fPrediction
    
    def set_prediction(self, value):
        self.fPrediction = value
    
    prediction = property(get_prediction, set_prediction)

    def get_jitter_radius(self):
        return self.fJitterRadius
    
    def set_jitter_radius(self, value):
        self.fJitterRadius = value
    
    jitter_radius = property(get_jitter_radius, set_jitter_radius)

    def get_max_deviation_radius(self):
        return self.fMaxDeviationRadius
    
    def set_max_deviation_radius(self, value):
        self.fMaxDeviationRadius = value
    
    max_deviation_radius = property(get_max_deviation_radius, set_max_deviation_radius)


#!/usr/bin/env python3

import sys
import os
from distutils.core import setup

from setuptools import find_packages
from glob import glob


data_files = glob("cvlab/images/*.*")+glob("cvlab/styles/*/*.*")+glob("cvlab/styles/*/*/*.*")
data_files = list(map(lambda x:x[6:], data_files))

if sys.version_info.major <= 2:
    raise Exception("Only python 3+ is supported!")

is_windows = os.name == 'nt'
is_linux = os.name == "posix"


requirements = [
    "numpy",
    "scipy",
    "pygments>=2",
    "matplotlib",
    "tinycss2",
]

try:
    import cv2
    if str(cv2.__version__) < "3":
        print("WARNING! OpenCV version 2.x detected. It is *strongly advised* to install OpenCV 3.x")
        print("Please visit: https://opencv.org/releases.html")
except ImportError:
    print("ERROR! OpenCV is required. Trying to use python-opencv package...")
    print("If it doesn't work, please visit: https://opencv.org/releases.html")
    requirements.append("python-opencv")


try:
    import PyQt5
except ImportError:
    requirements.append("pyqt5")


__version__ = None
package_name = None
exec(compile(open('cvlab/version.py').read(), 'cvlab/version.py', 'exec'))


this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

description = long_description.splitlines()[0].strip()

setup(
    name=package_name,
    version=__version__,
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Adam Brzeski, Jan Cychnerski',
    author_email='adam.m.brzeski@gmail.com, jan.cychnerski@gmail.com',
    url='https://github.com/cvlab-ai/cvlab',
    packages=find_packages(exclude=["diagrams","tools","samples","temp"]),
    package_data={"cvlab":data_files},
    entry_points={'gui_scripts': ['cvlab=cvlab.__main__:main']},
    license="AGPL-3.0+",
    python_requires='>=3.3',
    install_requires=requirements,
)



import threading
from collections import OrderedDict

import cv2 as cv

from . import id_manager
from .connectors import Output, Input
from .parameters import Parameter


class Element:
    name = "Unnamed element"
    comment = ""
    icon = None

    """
    Interface for all logic and GUI diagram objects.
    Methods must be thread-safe and non-blocking.
    Use of parameters must be thread-safe.
    """

    STATE_UNSET = 0
    STATE_BUSY = 1
    STATE_READY = 2
    STATE_ERROR = 3

    def __init__(self):
        super(Element, self).__init__()

        self.state = self.STATE_UNSET
        self.message = ""
        self.lock = threading.RLock()
        self.diagram = None
        self.object_id = id_manager.next_id(self)
        self.unique_id = str(id_manager.unique_id())

        inputs, outputs, parameters = self.get_attributes()

        self.inputs = OrderedDict()
        for i in inputs:
            assert isinstance(i, Input)
            i.parent = self
            self.inputs[i.id] = i

        self.outputs = OrderedDict()
        for i in outputs:
            assert isinstance(i, Output)
            i.parent = self
            self.outputs[i.id] = i

        self.parameters = OrderedDict()
        for i in parameters:
            assert isinstance(i, Parameter)
            #i.parent_element = self
            i.value_changed.connect(self.parameter_changed)
            self.parameters[i.id] = i

    def get_attributes(self):
        """
        Returns three lists: inputs, outputs and parameters
        """
        pass

    def parameter_changed(self):
        self.recalculate(True, False, True)

    #logic methods

    def recalculate(self, refresh_parameters, refresh_structure, force_break):
        """
        Informs that something has changed, and the element must re-do the calculations.
        If force_break=True, then it also forces element to break and cancel current calculations.
        If recal_structure=True, the it also recreates the structure of processing (necessary when Sequences are changed od elements are [dis]connected)
        """
        print("recalculate.")

    def delete(self):
        pass

    def may_interrupt(self):
        return False

    def set_state(self, state, info=""):
        if isinstance(info, cv.error):
            info = str(info)
            if "error: " in info:
                info = "OpenCV: " + info[info.index("error: ") + 7:]
        elif isinstance(info, Exception):
            info = info.__class__.__name__ + ": " + ", ".join(map(str, info.args))
        else:
            info = str(info)

        info = info.strip()

        with self.lock:
            if self.state == state and self.message == info: return
            self.state = state

            if state == self.STATE_READY:
                info = "Ready. " + info
            elif state == self.STATE_BUSY:
                info = "Waiting... " + info
            elif state == self.STATE_ERROR:
                info = "ERROR! " + info

            self.message = info

        self.notify_state_changed()

    def to_json(self):
        return {
            "_type": "element",
            "class": self.__class__.__name__,
            "module": self.__module__,
            "parameters": self.parameters,
            "unique_id": self.unique_id
        }

    def from_json(self, data):
        if "unique_id" in data:
            self.unique_id = data["unique_id"]
        for param, value in data["parameters"].items():
            if param in self.parameters:
                self.parameters[param].from_json(value)
        self.recalculate(True, True, True)


    #gui methods

    def notify_state_changed(self):
        """
        Informs the GUI that the processing is finished.
        """
        print("redraw.")



from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QApplication

from ..diagram import code_generator
from ..diagram.element import Element
from .parameters import *
from .mimedata import Mime
from .widgets import InOutConnector, ElementStatusBar, PreviewsContainer, StyledWidget
from .wires import NO_FOREGROUND_WIRES


SHOW_ELEMENT_ID = False


class GuiElement(Element, StyledWidget):
    state_changed = pyqtSignal()
    element_relocated = pyqtSignal(Element)

    def __init__(self):
        super(GuiElement, self).__init__()
        self.setObjectName("Element")
        self.setContextMenuPolicy(QtCore.Qt.ActionsContextMenu)
        self.status_bar = ElementStatusBar(self)
        self.label = None
        self.preview = None
        self.params = None
        self.input_connectors = {}
        self.output_connectors = {}
        self.param_sliders = []
        self.setAcceptDrops(True)
        self.hints_shown = False
        self.standard_actions = []
        self.group_actions = []
        self.workarea = None
        self.state_notified = False
        self.selected = False

    def set_selected(self, select):
        self.setProperty("selected", select)
        self.label.setProperty("selected", select)
        self.workarea.style().polish(self)
        self.workarea.style().polish(self.label)
        self.workarea.style().unpolish(self)
        self.workarea.style().unpolish(self.label)
        self.selected = select

    def set_workarea(self, workarea):
        self.workarea = workarea
        for connector in list(self.input_connectors.values()) + list(self.output_connectors.values()):
            connector.set_workarea(workarea)
        self.recreate_group_actions()

    def create_label(self, layout):
        if SHOW_ELEMENT_ID:
            self.label = QLabel("{} #{}".format(self.name, self.unique_id))
        else:
            self.label = QLabel("{}".format(self.name))
        self.label.setObjectName("ElementLabel")
        self.label.setMinimumWidth(100)  # workaround for drag problems - with pixmap width < 100 and adjusted HotSpot
        layout.addWidget(self.label)

    def create_params(self, container):
        self.params = container
        layout = container.layout()
        params = list(self.parameters.items())
        for name, param in params:
            if isinstance(param, ButtonParameter):
                layout.addLayout(GuiButtonParameter(param))
            elif isinstance(param, PathParameter):
                layout.addLayout(GuiPathParameter(param))
            elif isinstance(param, MultiPathParameter):
                layout.addLayout(GuiMultiPathParameter(param))
            elif isinstance(param, DirectoryParameter):
                layout.addLayout(GuiDirectoryParameter(param))
            elif isinstance(param, IntParameter):
                layout.addLayout(GuiIntParameter(param, self))
            elif isinstance(param, FloatParameter):
                layout.addLayout(GuiFloatParameter(param, self))
            elif isinstance(param, ComboboxParameter):
                layout.addLayout(GuiComboboxParameter(param, self))
            elif isinstance(param, SizeParameter) or isinstance(param, PointParameter):
                layout.addLayout(GuiMultiNumberParameter(param, self, 2, int))
            elif isinstance(param, ScalarParameter):
                layout.addLayout(GuiMultiNumberParameter(param, self, 4, float))
            elif isinstance(param, TextParameter):
                layout.addLayout(GuiTextParameter(param, self))

    def create_inputs(self, layout):
        layout.setContentsMargins(0, 3, 0, 3)
        layout.setSpacing(7)
        for input_ in self.inputs.values():
            is_input = True
            input_connector = InOutConnector(self, input_, is_input)
            layout.addWidget(input_connector)
            self.input_connectors[input_] = input_connector

    def create_outputs(self, layout):
        layout.setContentsMargins(0, 3, 0, 3)
        layout.setSpacing(7)
        for output in self.outputs.values():
            is_input = False
            output_connector = InOutConnector(self, output, is_input)
            layout.addWidget(output_connector)
            self.output_connectors[output] = output_connector

    def create_preview(self, layout):
        self.preview = PreviewsContainer(self, list(self.outputs.values()))
        layout.addWidget(self.preview)

    def create_switch_params_action(self):
        action = QAction('Show p&arams', self)
        action.triggered.connect(self.switch_params)
        action.setCheckable(True)
        self.standard_actions.append(action)
        self.addAction(action)

    def create_switch_preview_action(self):
        action = QAction('Show &preview\t[Double click]', self)
        action.triggered.connect(self.switch_preview)
        action.setCheckable(True)
        self.standard_actions.append(action)
        self.addAction(action)

    def create_switch_sliders_action(self):
        action = QAction('Show &sliders', self)
        action.triggered.connect(self.switch_sliders)
        action.setCheckable(True)
        self.standard_actions.append(action)
        self.addAction(action)

    def create_del_action(self):
        del_action = QAction('&Delete', self)
        del_action.triggered.connect(self.selfdestroy)
        self.standard_actions.append(del_action)
        self.addAction(del_action)

    def create_code_action(self):
        code_action = QAction('&Generate code', self)
        code_action.triggered.connect(self.gen_code_action)
        self.standard_actions.append(code_action)
        self.addAction(code_action)

    def create_duplicate_action(self):
        dup_action = QAction('D&uplicate', self)
        dup_action.triggered.connect(self.duplicate)
        self.standard_actions.append(dup_action)
        self.addAction(dup_action)

    def create_break_action(self):
        action = QAction('&Break connections', self)
        action.triggered.connect(self.break_connections)
        self.standard_actions.append(action)
        self.addAction(action)

    def create_menu_separator(self):
        separator = QAction(self)
        separator.setSeparator(True)
        self.standard_actions.append(separator)
        self.addAction(separator)

    def recreate_group_actions(self):
        self.group_actions[:] = []
        action = QAction('&Delete selected', self)
        action.triggered.connect(self.workarea.selection_manager.delete_selected)
        self.group_actions.append(action)
        self.addAction(action)

    def prepare_actions(self):
        group_selected = self.workarea.selection_manager.selected_count() > 1
        for action in self.standard_actions:
            action.setVisible(not group_selected)
        for action in self.group_actions:
            action.setVisible(group_selected)

    def update_id(self):
        if SHOW_ELEMENT_ID:
            self.label.setText("{} #{}".format(self.name, self.unique_id))

    @pyqtSlot()
    def gen_code_action(self):
        code = code_generator.generate(self)
        QApplication.instance().clipboard().setText(code)
        msg = QMessageBox(QMessageBox.Information, "Code copied",
                                "Code copied to system clipboard.\n\n" + code[:200] + "...", QMessageBox.Ok, self)
        msg.setModal(True)
        msg.show()

    @pyqtSlot()
    def switch_params(self, value=None):
        if not self.params: return
        if value is None: value = not self.params.isVisible()
        self.params.setVisible(value)
        for action in self.actions():
            if action.text() == 'Show p&arams':
                action.setChecked(value)

    @pyqtSlot()
    def switch_preview(self, value=None):
        self.preview.switch_visibility(value)
        for action in self.actions():
            if action.text() == 'Show &preview':
                action.setChecked(self.preview.isVisible())

    @pyqtSlot()
    def switch_sliders(self, value=None):
        if not self.param_sliders:
            return
        if value is None: value = not self.param_sliders[0].isVisible()
        for slider in self.param_sliders:
            slider.setVisible(value)
        for action in self.actions():
            if action.text() == 'Show &sliders':
                action.setChecked(value)

    @pyqtSlot()
    def break_connections(self):
        if not self.parameters: return
        for par in list(self.parameters.values()):
            par.disconnect_all_children()

    @pyqtSlot()
    def selfdestroy(self):
        #TODO: do we need this here or somewhere else?
        self.diagram.delete_element(self)

    def mousePressEvent(self, e):
        if e.button() == QtCore.Qt.MiddleButton:
            e.ignore()
            return
        self.workarea.on_element_moused_pressed(self, e)
        if e.button() == QtCore.Qt.RightButton:
            self.prepare_actions()
        self.raise_()
        if not NO_FOREGROUND_WIRES:
            self.workarea.wires_in_foreground.raise_()
        # see: WiresBase.__init__ comments!
        # self.setGraphicsEffect(QGraphicsOpacityEffect())
        self.hide_hints()

    def mouseDoubleClickEvent(self, e):
        if e.button() == QtCore.Qt.MiddleButton:
            e.ignore()
            return
        self.switch_preview()

    def mouseReleaseEvent(self, e):
        if e.button() == QtCore.Qt.MiddleButton:
            e.ignore()
            return
        if e.button() == QtCore.Qt.LeftButton:
            self.workarea.on_element_moused_left_released(self, e)
        # self.setGraphicsEffect(None)
        self.show_hints()

    def mouseMoveEvent(self, e):
        if e.buttons() & QtCore.Qt.MiddleButton:
            e.ignore()
            return
        if e.buttons() & QtCore.Qt.LeftButton:
            self.workarea.on_element_moused_left_moved(self, e)
            #self.show_hints()

    def show_hints(self):
        if self.hints_shown: return
        self.hints_shown = True
        for i in list(self.input_connectors.values()) + list(self.output_connectors.values()):
            i.show_hint()

    def hide_hints(self):
        if not self.hints_shown: return
        self.hints_shown = False
        for i in list(self.input_connectors.values()) + list(self.output_connectors.values()):
            i.hide_hint()

    def enterEvent(self, e):
        self.show_hints()

    def leaveEvent(self, e):
        if not self.workarea.wires_in_foreground.is_cursor_line_active:
            self.hide_hints()

    def forward_event_to_io_connector(self, name, event):
        if event.mimeData().text() == Mime.INCOMING_CONNECTION and self.input_connectors:
            connector = self.find_nearest_connector(event, list(self.input_connectors.values()))
            method_to_call = getattr(connector, name)
            method_to_call(event)
        elif event.mimeData().text() == Mime.OUTGOING_CONNECTION and self.output_connectors:
            connector = self.find_nearest_connector(event, list(self.output_connectors.values()))
            method_to_call = getattr(connector, name)
            method_to_call(event)

    def find_nearest_connector(self, event, connectors):
        key = lambda c: abs(event.pos().y() - c.pos().y() - c.height() // 2)
        return min(connectors, key=key)

    def resizeEvent(self, e):
        self.element_relocated.emit(self)

    def dragEnterEvent(self, e):
        self.forward_event_to_io_connector(self.dragEnterEvent.__name__, e)

    def dropEvent(self, e):
        self.forward_event_to_io_connector(self.dropEvent.__name__, e)

    def dragMoveEvent(self, e):
        #todo: fix shifted wire when to connectors of the same element are to be connected
        self.forward_event_to_io_connector(self.dragMoveEvent.__name__, e)

    def notify_state_changed(self):
        if not self.state_notified:
            self.state_notified = True
            # We need to use a signal here since the method is called by a worker thread and we need to alter the GUI
            self.state_changed.emit()

    def duplicate(self):
        el = self.__class__()
        pos = (self.pos().x() + 20, self.pos().y() + 20)
        self.diagram.add_element(el, pos)
        for my, his in zip(list(self.parameters.values()), list(el.parameters.values())):
            my.connect_child(his)
            his.connect_child(my)
        for my, his in zip(list(self.inputs.values()), list(el.inputs.values())):
            for outp in my.connected_from:
                self.diagram.connect_io(his, outp)
        if self.params:
            el.switch_params(self.params.isVisible())
        if self.param_sliders:
            el.switch_sliders(self.param_sliders[0].isVisible())
        el.switch_preview(self.preview.isVisible())

    def to_json(self):
        parent_d = Element.to_json(self)
        d = {
            "show_parameters": (self.params.isVisible() if self.params else None),
            "show_sliders": (self.param_sliders[0].isVisible() if self.param_sliders else None),
            "show_preview": (self.preview.isVisible() if self.preview else None),
            "position": (self.pos().x(), self.pos().y()),
            "preview_size": self.preview.preview_size,
        }
        parent_d["gui_options"] = d
        return parent_d

    def from_json(self, data):
        options = data["gui_options"]
        self.switch_params(options['show_parameters'] is True)
        self.switch_sliders(options["show_sliders"])
        if "preview_size" in options \
            and options["preview_size"] \
            and options["preview_size"] != self.preview.preview_size:
                self.preview.preview_size = options["preview_size"]
        self.switch_preview(options["show_preview"])
        self.move(options["position"][0], options["position"][1])
        Element.from_json(self, data)
        self.update_id()
        self.preview.force_update()

    def zoom(self, factor, origin):
        assert isinstance(self.preview, PreviewsContainer)

        # fixme: this is not accurate (elements are zoomed by top-left position, sizes and positions are integers...)

        factor = float(factor)
        origin_x, origin_y = origin

        x, y = self.workarea.nearest_grid_point((self.x() - origin_x) * factor + origin_x,
                                                (self.y() - origin_y) * factor + origin_y)
        self.move(x, y)

        self.preview.resize_previews(self.preview.preview_size * factor)

        self.element_relocated.emit(self)

    def deleteLater(self):
        self.state_changed.disconnect()
        # self.setParent(None)
        super(GuiElement, self).deleteLater()


class FunctionGuiElement(GuiElement):
    def __init__(self):
        super(FunctionGuiElement, self).__init__()
        vb_main = QVBoxLayout()
        hb_content = QHBoxLayout()
        hb_label = QHBoxLayout()
        vb_inputs = QVBoxLayout()
        vb_params = QVBoxLayout()
        vb_outputs = QVBoxLayout()
        vb_inputs.setAlignment(QtCore.Qt.AlignTop)
        vb_outputs.setAlignment(QtCore.Qt.AlignTop)

        w_params = QWidget()
        w_params.setLayout(vb_params)
        vb_params.setContentsMargins(0,0,0,0)

        self.create_label(hb_label)
        self.create_params(w_params)
        self.create_inputs(vb_inputs)
        self.create_outputs(vb_outputs)
        hb_content.addLayout(vb_inputs)
        hb_content.addWidget(w_params)
        hb_content.addStretch(1)
        hb_content.addLayout(vb_outputs)
        vb_main.addLayout(hb_label)
        vb_main.addLayout(hb_content)
        vb_main.addWidget(self.status_bar)
        vb_main.setSizeConstraint(QLayout.SetFixedSize)
        vb_main.setContentsMargins(3, 3, 3, 3)
        self.setLayout(vb_main)

        self.create_preview(vb_main)
        self.create_switch_params_action()
        self.create_switch_preview_action()
        self.create_switch_sliders_action()
        self.create_menu_separator()
        self.create_duplicate_action()
        self.create_break_action()
        self.create_del_action()
        self.create_code_action()
        #self.setFocusPolicy(QtCore.Qt.ClickFocus + QtCore.Qt.TabFocus)
        #self.setAttribute(QtCore.Qt.WA_MacShowFocusRect, 1)     # enable showing focus on a Mac


class OperatorGuiElement(GuiElement):
    def __init__(self):
        super(OperatorGuiElement, self).__init__()
        vb_main = QVBoxLayout()
        hb = QHBoxLayout()
        vb_inputs = QVBoxLayout()
        vb_outputs = QVBoxLayout()
        vb_inputs.setAlignment(QtCore.Qt.AlignTop)
        vb_outputs.setAlignment(QtCore.Qt.AlignTop)

        self.create_label(vb_main)
        self.create_inputs(vb_inputs)
        self.create_outputs(vb_outputs)
        hb.addLayout(vb_inputs)
        hb.addStretch(1)
        hb.addLayout(vb_outputs)
        vb_main.addLayout(hb)
        vb_main.addWidget(self.status_bar)
        vb_main.setSizeConstraint(QLayout.SetFixedSize)
        vb_main.setContentsMargins(3, 3, 3, 3)
        self.create_preview(vb_main)
        self.setLayout(vb_main)

        self.create_switch_preview_action()
        self.create_menu_separator()
        self.create_del_action()
        self.create_code_action()
        self.create_duplicate_action()


class InputGuiElement(GuiElement):
    def __init__(self):
        super(InputGuiElement, self).__init__()
        vb_main = QVBoxLayout()
        hb = QHBoxLayout()
        vb_params = QVBoxLayout()
        vb_outputs = QVBoxLayout()
        vb_outputs.setAlignment(QtCore.Qt.AlignTop)

        w_params = QWidget()
        w_params.setLayout(vb_params)
        vb_params.setContentsMargins(0,0,0,0)

        self.create_label(vb_main)
        self.create_params(w_params)
        self.create_outputs(vb_outputs)
        hb.addWidget(w_params)
        hb.addLayout(vb_outputs)
        vb_main.addLayout(hb)
        vb_main.addWidget(self.status_bar)
        vb_main.setSizeConstraint(QLayout.SetFixedSize)
        vb_main.setContentsMargins(3, 3, 3, 3)
        self.create_preview(vb_main)
        self.setLayout(vb_main)

        self.create_switch_preview_action()
        self.create_menu_separator()
        self.create_duplicate_action()
        self.create_break_action()
        self.create_del_action()

